import pandas as pd
import muldataframe.cmm as cmm
import muldataframe as md
# import muldataframe.cmm as cmm
# import typing

def aggregate_index(i:int,index:pd.DataFrame,index_agg:cmm.IndexAgg) -> pd.DataFrame:
    # agg_mode: 'same_only', 'array'
    # print(key)
    
    # if isinstance(key,tuple) or \
    #     not isinstance(key,typing.Hashable):
    #     final_index = pd.Index([i])
    # else:
    #     final_index = pd.Index([key])
    final_index = pd.Index([i])
    if index_agg == 'same_only':
        # print(index[index.columns[0]].unique())
        index_same = index.loc[:,[len(index[col].unique())==1 for col in index.columns]]
        index_one = index_same.iloc[[0]]
        index_one.index = final_index
        # print('====',index,'\n',index_same,'\n',index_one)
        return index_one
    elif index_agg in ['list','tuple']:
        index_vals = []
        for col in index.columns:
            vals = index[col].unique()
            if len(vals) == 1:
                index_vals.append(vals[0])
            else:
                vals = tuple(vals) if index_agg == 'tuple' else vals
                index_vals.append(vals)
        return pd.DataFrame([index_vals],columns=index.columns,index=final_index)

        

def concat(arr:list[md.MulSeries|md.MulDataFrame],axis=0):
    '''
    Concatenate muldataframe objects along a particular axis.

    The list of muldataframe objects are joined by the primary index or columns depending on the specified axis. Index or columns dataframes along the axis are concatenated. Index or columns dataframes on the other axis are ignored except for the first one. The first one is used as the index or columns dataframe of the final muldataframe object. Currently, only inner join is supported. 

    Parameters
    -------------
    objs : a list of MulSeries or MulDataFrame objects
        It can be a mix of mulseries and muldataframes.
    axis : {0, 1}, default 0
        The axis to concatenate along.

    Returns
    ----------
    MulSeries or MulDataFrame
        It will return a mulseries only if the objects in the list are all mulseries and are concatenated along ``axis=0``
    
    Examples
    ---------
    >>> import pandas as pd
    >>> import muldataframe as md
    >>> index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','c'],
                     columns=['x','y'])
    >>> columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    >>> mf = md.MulDataFrame([[1,2],[8,9],[8,10]],index=index, columns=columns)
    >>> ms = md.MulSeries([5,6,7],
            index=pd.DataFrame(index=['a','b','c']),
            name=pd.Series([8,9],index=['g','f']))
    >>> pd.concat([mf,ms],axis=1)
    (3, 3)    g  7   6  8
              f  5   3  9
                 c   d  k
    --------  ------------
       x  y      c   d  k
    a  1  2   a  1   2  5
    b  3  6   b  8   9  6
    c  5  6   c  8  10  7
    >>> pd.concat([ms,mf],axis=1)
    (3, 3)            f  9  5   3
                      g  8  7   6
                         k  c   d
    ----------------  ------------
    Empty DataFrame      k  c   d
    Columns: []       a  5  1   2
    Index: [a, b, c]  b  6  8   9
                      c  7  8  10
    '''
    ds_new = pd.concat([x.ds for x in arr],join='inner',axis=axis)
   
    # print('ddddddd',mds1.ds,mds2.ds,ds_new)
    if isinstance(ds_new,pd.Series):
        mindex_new = pd.concat([x.index for x in arr],join='inner')
        return md.MulSeries(ds_new.values,index=mindex_new,
                    name=arr[0].name,index_copy=False)
    else:
        if axis == 0:
            mindex_new = pd.concat([x.index for x in arr],join='inner')
            return md.MulDataFrame(ds_new.values,index=mindex_new,
                    columns=arr[0].columns,index_copy=False)
        else:
            mcols_new = pd.concat([x.columns 
                if isinstance(x,md.MulDataFrame) else pd.DataFrame(x.name).transpose() for x in arr],join='inner')
            # print('***************')
            # print(ds_new)
            # print(mcols_new)
            return md.MulDataFrame(ds_new.values,index=arr[0].index,
                    columns=mcols_new,columns_copy=False)
    

def pivot_table(*args,**kwargs):
    '''
    Same as ``pandas.pivot_table`` except that it returns a MulDataFrame.

    If with the same arguments, ``pandas.pivot_table`` returns a dataframe with no multiindex. Then this function will return a muldataframe with empty index and columns dataframes. 
    
    Check :doc:`MulDataFrame.melt <../muldataframe/melt>` for a reverse operation.

    Parameters
    -----------
    The parameters are exactly the same as in `pandas.pivot_table <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.pivot_table.html#pandas.DataFrame.pivot_table>`_. Please check its web page for detailed usage.

    Returns
    --------
    MulDataFrame
        returns a MulDataFrame

    Examples
    ----------
    >>> import pandas as pd
    >>> import muldataframe as md
    >>> df = pd.DataFrame({"A": ["foo", "foo", "foo", "foo", "foo",
                         "bar", "bar", "bar", "bar"],
                   "B": ["one", "one", "one", "two", "two",
                         "one", "one", "two", "two"],
                   "C": ["small", "large", "large", "small",
                         "small", "large", "small", "small",
                         "large"],
                   "D": [1, 2, 2, 3, 3, 4, 5, 6, 7],
                   "E": [2, 4, 5, 5, 6, 6, 8, 9, 9]})
    >>> md.pivot_table(df, values='D', index=['A', 'B'], columns=['C'], aggfunc="sum")
    (4, 2)       Empty DataFrame
                 Columns: []
                 Index: [large, small]
    -----------  -----------------------
        A    B   C  large  small
    0  bar  one  0    4.0    5.0
    1  bar  two  1    7.0    6.0
    2  foo  one  2    4.0    1.0
    3  foo  two  3    NaN    6.0
    '''
    df = pd.pivot_table(*args,**kwargs)
    # print(df)
    if isinstance(df.index,pd.MultiIndex):
        new_idx = df.index.to_frame(index=False)
    else:
        new_idx = pd.DataFrame(index=df.index)
    if isinstance(df.columns,pd.MultiIndex):
        new_cols = df.columns.to_frame(index=False)
    else:
        new_cols = pd.DataFrame(index=df.columns)
    
    return md.MulDataFrame(df.values,
                           index=new_idx,
                           columns=new_cols,
                           both_copy=False)


import pandas as pd
import muldataframe.cmm as cmm
# import muldataframe.util as util
from typing import Any
import numpy as np
import muldataframe as md
# import muldataframe.ValFrameBase as vfb
import muldataframe.ValFrameBase as vfb
import tabulate
tabulate.PRESERVE_WHITESPACE = True

# import collections.abc

#TODO: query for mulseries and muldataframe

class MulSeries:
    '''
    A multi-index series with the index being a pandas dataframe and the name a pandas series. It also has an underlying values series that is not directly accessible. Its values are the same as the values of the values series.

    Parameters
    -----------
    data: pandas.Series, array-like, Iterable, dict, or scalar value
        either a pandas Series or the same kind of data argument as required in the `pandas Series constructor <https://pandas.pydata.org/docs/reference/api/pandas.Series.html#pandas.Series>`_. The values series is constructed from the data argument.
    index: pandas.DataFrame
        If index is None, construct an empty index dataframe using the index of the values series as its index.
    name: pandas.Series, str
        If name is of str type, construct an empty name series using name as its name. If name is None, construct an empty name series using the name of the values series as its name.
    index_init: Literal['overlap'] | Literal['override'] | Literal['align']
        The option determins how to align the index of the index dataframe to the index of the values series. 
        
        - 'overlap' : The overlap of the index of the index dataframe and the index of the values series is used to index the values series and the index dataframe. This mode requires both indexes being unique. 
        - 'override' : the index of the index dataframe overrides the index of the values series. This mode requires both indexes' lengths to be the same. 
        - 'align' : This mode is only effective if the data argumnet implies an index and the index argument is not None. In this mode, the index of the index dataframe is used to index the values series constructed from the data argument. The resulting series is used as the final values series. It requires the index of the values series being uinque and have all the labels in the index dataframe's index. 
        - None : the default behavior depends on the type of the ``data`` argument. If ``data`` implies a row index such as if it is a ``pandas.Series`` object or a dict, the mode will be ``'overlap```. Otherwise, it will be ``'override'``.
        
        Of note, when ``data`` implies a row index and this index is the same as the index of the index dataframe,  the ``'override'`` mode will be used even if other modes are specified. In all three modes, the name of the initialized primary index will be the same as the name of the index of the index dataframe.
    index_copy: bool
        whether to create a copy of the index argument.
    name_copy: bool
        whether to create a copy of the name argument.
    data_copy : bool, default None
        Wether to copy ``data``. It behaves the same as the ``copy`` argument in `pandas.Series.__init__ <https://pandas.pydata.org/docs/reference/api/pandas.Series.html>`_

    Examples:
    ----------
    Construct a mulseries. Notice that the index of the dataframe and the index of the values series are the same and the name of the name series and the name of the values series are the same.
    
    >>> import pandas as pd
    >>> import muldataframe as md
    >>> index = pd.DataFrame([[1,2],[3,5],[3,6]],
                            index=['a','b','b'],
                            columns=['x','y'])
    >>> name = pd.Series(['g','h'],index=['e','f'], name='cc')
    >>> ms = md.MulSeries([1,2,3],index=index,name=name)
    >>> ms
    (3,)     f   h
             e   g
                cc
    -------  ------
       x  y     cc
    a  1  2  a   1
    b  3  5  b   2
    b  3  6  b   3
    '''
    # force pandas to return NotImplemented when using ops like +, * 
    # in the case of pd.Series + MulSeries.
    __pandas_priority__ = 10000
    def __init__(self,data,index:pd.DataFrame=None,
                 name:pd.Series|str|None=None,
                 index_init:cmm.IndexInit=None,
                 index_copy=True,name_copy=True,
                 data_copy=None):
       
        ss = data
        
        if isinstance(ss,dict):
            ss = pd.Series(ss,copy=data_copy)

        if isinstance(ss,pd.Series):
            index_init = 'overlap' if index_init is None else index_init
            if data_copy:
                ss = ss.copy()
        else:
            index_init = 'override' if index_init is None else index_init
            ss = pd.Series(ss,copy=data_copy)

        if not isinstance(name,pd.Series):
            if isinstance(name,str):
                name = pd.Series([],name=name)
            else:
                name = pd.Series([],name=ss.name)
        else:
            name = name.copy() if name_copy else name

        ss, index = cmm.setMulIndex(ss,'index',index,index_init,index_copy)

        
        self.index = index
        '''
        The index dataframe. 

        Use :doc:`MulSeries.mindex <mindex>` as an alias for this attribute.
        '''
        self.name = name
        '''
        The name series. 

        Use :doc:`MulSeries.mname <mname>` as an alias for this attribute.
        '''
        # print(hasattr(self, 'index'))
        self.__ss = ValSeries(self,ss) # private

        self.iloc = cmm.Accessor(self._xloc_get_factory('iloc'),
                             self._xloc_set_factory('iloc'))
        '''
        Position-based indexing. 

        It is the same as the `Series.iloc <https://pandas.pydata.org/docs/reference/api/pandas.Series.iloc.html>`_ of the values series except that it returns a MulSeries with the index dataframe properly sliced. If the return value is a scalar, it returns the scalar.
        '''
        self.loc = cmm.Accessor(self._xloc_get_factory('loc'),
                            self._xloc_set_factory('loc'))
        '''
        Label-based indexing. 
        
        It is the same as `Series.loc <https://pandas.pydata.org/docs/reference/api/pandas.Series.loc.html>`_ of the values series except that it returns a MulSeries with the index dataframe properly sliced. If the return value is a scalar, it returns the scalar.
        '''
        self.mloc = cmm.Accessor(self._mloc_get,
                             self._mloc_set)
        '''
        Flexible hierachical indexing on the index dataframe. 
        
        The slicer can be a list or a dict. Check introduction to mloc ??? for detailed usage. 
        
        If a list is used, its length should be less than or equal to the columns length of the index dataframe. The hierarchical indexing order is from the leftmost column to the rightmost. Use ``...`` as ``:`` in the list to select all elements in a column.

        If a dict is used, its keys should be the column names of the index dataframe and its values the slicers on the columns. The hierachical indexing order is the insertion order of the keys in the dict. Although Python does not guanrantee the insertion order, it is preserved in most cases. Use the `OrderedDict <https://docs.python.org/3/library/collections.html#collections.OrderedDict>`_ class if you are really concerned about it.

        Examples
        ---------
        List indexing:

        >>> import pandas as pd
        >>> import muldataframe as md
        >>> index = pd.DataFrame([['a','b','c'],
                                  [ 'g','b','f'],
                                  [ 'b','g','h']],
                           columns=['x','y','y'])
        >>> name = pd.Series(['a','b'],index=['e','f'],name='cc')
        >>> ms = md.MulSeries([1,2,3],index=index,name=name)
        >>> ms.mloc[[..., 'b']]
        (2,)        e   a
                    f   b
                       cc
        ----------  ------
           x  y  y     cc
        0  a  b  c  0   1
        1  g  b  f  1   2
        >>> ms.mloc[['g', ..., ['h','f']]]
        2

        Dictionary indexing:

        >>> ms.mloc[{'y':['c','h'],'x':['b','a']}]
        (2,)        e   a
                    f   b
                       cc
        ----------  ------
           x  y  y     cc
        2  b  g  h  2   3
        0  a  b  c  0   1

        Note in the above example that if the index dataframe's columns have duplicate names, use the **last** column for dict indexing.
        
        '''
        self.nloc = cmm.Accessor(self._nloc_get,
                             self._nloc_set)
        '''
        Flexible hierachical indexing on the index dataframe using positions. 
        
        The slicer can be a list or a dict. Check introduction to mloc ??? for detailed usage. 
        
        If a list is used, it behaves exactly like :doc:`mloc`.

        If a dict is used, it behaves similarly to :doc:`mloc` except that instead of using column names as keys, it uses the numeric positions of the columns as keys.

        Examples
        ---------
        Dictionary indexing:

        >>> import pandas as pd
        >>> import muldataframe as md
        >>> index = pd.DataFrame([['a','b','c'],
                                  [ 'g','b','f'],
                                  [ 'b','g','h']],
                           columns=['x','y','y'])
        >>> name = pd.Series(['a','b'],index=['e','f'],name='cc')
        >>> ms = md.MulSeries([1,2,3],index=index,name=name)
        >>> ms.nloc[{1:['b','g'],0:['b','a']}]
        (2,)        e   a
                    f   b
                       cc
        ----------  ------
           x  y  y     cc
        2  b  g  h  2   3
        0  a  b  c  0   1

        Note that with a dict in :doc:`mloc`, you can only select the last ``y`` column in the index dataframe. Using ``nloc`` you are able to select the first ``y`` column.
        '''

    def __repr__(self):
        cols = cmm.fmtSeries(self.name)
        vals = cmm.fmtSeries(self.ds)
        return tabulate.tabulate(
                [[str(self.index),str(vals)]],
               headers=[self.shape,
                        cmm.fmtColStr(cols,False)])
        # return 'ss:\n'+self.__ss.__repr__()+'\n\nindex:\n'+\
        #         self.index.__repr__()+'\n\nname:\n'+\
        #         self.name.__repr__()
    
    def __getattr__(self,name):
        if name == 'values':
            return self.__ss.values
        elif name == 'ss':
            return pd.Series(self.values.copy(),
                             index=self.index.index.copy(),
                             name=self.name.name)
        elif name in ['mindex','mname']:
            return getattr(self,name.lstrip('m'))
        elif name == 'midx':
            return self.index
        elif name in ['primary_index','pindex','pidx']:
            return self.index.index
        elif name in ['primary_name','pname']:
            return self.name.name
        elif name == 'shape':
            return self.__ss.shape
        elif name == 'ds':
            # values are not copied version
            return pd.Series(self.values,
                             index=self.index.index.copy(),
                             name=self.name.name,
                             copy=False)
        elif cmm.is_pandas_method(self,name):
            def func(*args,**kwargs):
                return self.call(name,*args,**kwargs)
            return func
        elif cmm.is_numpy_function(name):
            def func(*args,**kwargs):
                return self.call(getattr(np,name),*args,**kwargs)
            return func

        
    def _hasVal(self):
        return self.__ss is not None

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ['ss','ds']:
            raise AttributeError(f"ss is a read-only attribute.")
        elif name in ['index','mindex','midx']:
            cmm.checkSetIdxValue(self,'index',value)
            super().__setattr__('index', value)
        elif name in ['manme']:
            super().__setattr__('name', value)
        elif name in ['primary_index','pindex','pidx']:
            self.index.index = value
        elif name in ['primary_name','pname']:
            self.name.name = value
        else:
            super().__setattr__(name, value)

    def __len__(self):
        return self.shape[0]
    
    def insert(self,label,value,loc=None,name=None,inplace=True):
        '''
        Insert a value into MulSeries at specified location.

        Duplicate labels are allowed. It is an append operation if ``loc=None``.

        Parameters
        ------------
        label : str, number, or hashable object
            Label of the inserted value
        value : Scalar
            The inserted value.
        loc : int
            Insertion index. Must verify 0 <= loc <= MulSeries.shape[0] if ``axis==0``. If ``loc=None``, insert at the end of the MulSeries, namely an append operation.
        name : Scalar, Series, or array-like
            The metadata of ``value`` inserted into the index dataframe.
        inplace : bool, default True
            Whether insert inplace or return a new MulSeries.

        Returns
        ---------
        None or MulSeries
            returns ``None`` if ``inplace=True``
        
        Examples
        ---------
        >>> import pandas as pd
        >>> import muldataframe as md
        >>> index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
        >>> columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
        >>> ms = MulSeries([1,8,9],index=index,name=columns.loc['c'])
        >>> ms.insert('e',7,loc=1,name=[8,9],inplace=False)
        (4,)     g  7
                 f  5
                    c
        -------  ------
           x  y     c
        a  1  2  a  1
        e  8  9  e  7
        b  3  6  b  8
        b  5  6  b  9
        '''
        loc = self.shape[0] if loc is None else loc
        if inplace:
            # self.__ss._update_super_index()
            ds = self.ds
            self.__ss = None
            dsf = pd.DataFrame(ds,copy=False).transpose()
            dsf.insert(loc,label,value)
            
            mindext = self.mindex.transpose()
            if name is None:
                name = [None]*self.mindex.shape[1]
            mindext.insert(loc,label,name)
            self.mindex = mindext.transpose()
            # print('\t',mindext)
            # print('\t',self.index)

            self.__ss = dsf.iloc[0]
        else:
            ms = self.copy()
            ms.insert(label,value,loc,name,True)
            return ms

    
    def drop(self,labels,mloc=None,inplace=False):
        '''
        Remove values from MulSeres by labels.

        Parameters
        ------------
        labels : single label or list-like
            primary index labels if ``mloc=None`` or labels in a column of the index dataframe specified by ``mloc``.
        mloc : None, str, number, or hashable object
            Column name in the index dataframe. if ``mloc=None``, use the primary index to select removed values. Otherwise, use the labels in the specified column to select removed values.
        inplace : bool, default False
            Whether to modify the MulSeries inplace or return a new MulSeries with values removed.
        
        Returns
        --------
        MulSeries or None
            returns ``None`` if ``inplace=True``.
        
        Examples
        -----------
        >>> import pandas as pd
        >>> import muldataframe as md
        >>> index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
        >>> columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
        >>> ms = MulSeries([1,8,9],index=index,name=columns.loc['c'])
        >>> ms.drop('b')
        (1,)     g  7
                 f  5
                    c
        -------  ------
           x  y     c
        a  1  2  a  1
        >>> ms.drop(6,mloc='y')
        (1,)     g  7
                 f  5
                    c
        -------  ------
           x  y     c
        a  1  2  a  1
        '''
        if inplace:
            self.__ss._update_super_index()
            __ss = self.__ss
            self.__ss = None
            if mloc is None:
                __ss.drop(labels,inplace=True)
                self.mindex.drop(labels,inplace=True)
            else:
                __ss.index = self.mindex[mloc]
                self.mindex['_&%@x'] = self.mindex.index
                self.mindex.index = self.mindex[mloc]
                self.mindex.drop(labels,inplace=True)
                self.mindex.index = self.mindex['_&%@x']
                self.mindex.drop('_&%@x',inplace=True,axis=1)
                __ss.drop(labels,inplace=True)
                __ss.index = self.mindex.index
            self.__ss = __ss
        else:
            ms = self.copy()
            ms.drop(labels,mloc,inplace=True)
            return ms

   
    def equals(self,other):
        '''
        Test whether two MulSeries are the same.

        Two MulSeries are equal only if their index dataframes, name series and value dataframes are equal. Use `Series.equals <https://pandas.pydata.org/docs/reference/api/pandas.Series.equals.html>`_ and `DataFrame.equals <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.equals.html#>`_ under the hood.

        Parameters
        ------------
        other : object
            The other object to be compared with the MulSeries. If the other is not a MulSeries, returns False.

        Returns
        ----------
        bool
            True for equality.
        '''
        if not isinstance(other,MulSeries):
            return False
        else:
            return self.ds.equals(other.ds) and self.index.equals(other.index) and self.name.equals(other.name)

    def copy(self,data_copy=True):
        '''
        Create a deep copy of MulSeries.

        Parameters
        ------------
        data_copy : bool, default True
            Whether to create a deep copy of the ``.values`` attribute.
        
        Returns
        ------------
        MulSeries
            A copied mulseries.
        '''
        if data_copy:
            return MulSeries(self.__ss.copy().values,
                         index=self.index,
                         name=self.name.copy())
        else:
            return MulSeries(self.values,
                         index=self.index,
                         name=self.name,
                         index_copy=True,
                         name_copy=True,
                         data_copy=False)
    
    def __iter__(self):
        '''
        Return an iterator of the values.

        Use `Series.__iter__ <https://pandas.pydata.org/docs/reference/api/pandas.Series.__iter__.html#pandas.Series.__iter__>`_ of the values series under the hood. 
        '''
        return self.__ss.__iter__()


    def __getitem__(self,key):
        if isinstance(key,MulSeries):
            key = key.ds
        new_ss = self.__ss[key]
        if(isinstance(new_ss,pd.Series)):
            idx_new = self.mindex.loc[key]
                # print('ok')
            ms = MulSeries(new_ss,
                                index=idx_new,
                                name=self.name,
                                index_init='override')
            return ms
        else:
            return new_ss
    
    def __setitem__(self,key, values):
        if isinstance(key,MulSeries):
            key = key.ds
        self.__ss[key] = values
        if not cmm.array_like(key) and key not in self.pindex:
            self.index.loc[key] = [None]*self.index.shape[1]
    
    def _xloc_get_factory(self,attr):
        def _xloc_get(key):
            if isinstance(key,MulSeries):
                key = key.ds
            new_ss = getattr(self.__ss,attr)[key]
            if(isinstance(new_ss,pd.Series)):
                idx_new = getattr(self.mindex,attr)[key]
                ms = MulSeries(new_ss,
                                    index=idx_new,
                                    name=self.name,
                                    index_init='override')
                return ms
            else:
                return new_ss
        return _xloc_get
    
    def _xloc_set_factory(self,attr):
        def _xloc_set(key,values):
            if isinstance(key,MulSeries):
                key = key.ds
            getattr(self.__ss,attr)[key] = values
            if attr == 'loc' and not cmm.array_like(key) and key not in self.pindex:
                self.index.loc[key] = [None]*self.index.shape[1]
        return _xloc_set

    def _mloc2pos(self,key):
        nx = cmm._mloc_idx(key,self.index)
        return nx
    
    def _mloc_get(self,key):
        if key == slice(None):
            return self.iloc[:]
        else:
            nx = self._mloc2pos(key)
            # print(nx)
            return self.iloc[nx]
    
    def _mloc_set(self,key,values):
        if key == slice(None):
            self.iloc[:] = values
        else:
            nx = self._mloc2pos(key)
            self.iloc[nx] = values
    
    def _nloc2pos(self,key:dict):
        return cmm._nloc_idx(key,self.index)
    
    def _nloc_get(self,key):
        if isinstance(key,dict):
            nx = self._nloc2pos(key)
            return self.iloc[nx]
        else:
            return self._mloc_get(key)
    
    def _nloc_set(self,key,values):
        if isinstance(key,dict):
            nx = self._nloc2pos(key)
            self.iloc[nx] = values
        else:
            self._mloc_set(key,values)


    def __fill_cols(self,col_fill,cols,inplace):
        if isinstance(col_fill,pd.Series) or \
            isinstance(col_fill,pd.DataFrame):
            if isinstance(col_fill,pd.Series):
                col_fill = pd.DataFrame(col_fill).transpose()
            col_fill = cmm.test_idx_eq(col_fill,cols,copy=False)

            col_fill = cmm.test_idx_eq(col_fill,self.name.index,indexType='columns',copy=False)
            
            mcols = pd.DataFrame(self.name).transpose()
            mcols = pd.concat([col_fill,mcols],axis=0)
            return mcols
        else:
            mcols = pd.DataFrame(self.name)
            for i, col in enumerate(cols):
                mcols.insert(i,col,col_fill,allow_duplicates=True)
            return mcols.transpose()
            
    def reset_index(self,columns=None, drop=False, 
                    inplace=False, col_fill=''):
        '''
        Reset the columns of the index dataframe as the columns of the mulseries.

        Parameters
        ----------
        columns : column name(s) of the index dataframe.
            If this argument is None, reset the index of the index dataframe. If the name of this index is None, it will be named as "primary_index". If "primary_index" exists in the primary columns, it will be named as "primary_index_1" and so on.
        drop : bool, default False
            Just reset the index, without inserting index dataframe's column(s) as column(s) in the new MulDataFrame.
        inplace : bool, default False
            Modify the MulSeries in place (do not create a new object).
        col_fill : object, default ''
            A scalar, a pandas Series or a pandas DataFrame to fill in the columns dataframe of the new MulDataFrame for the inserted values. If the argument is a Series or a DataFrame, its index should align with the index of the mulseries' name (which is a pandas series) in the same way as the align mode in the :doc:`constructor <mulseries>`.

        Returns
        --------
        MulSeries, MulDataFrame or None
            The return value behaves similarly to `Series.reset_index <https://pandas.pydata.org/docs/reference/api/pandas.Series.reset_index.html>`_.
        
        Examples
        ---------
        >>> import muldataframe as md
        >>> index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
        >>> name = pd.Series([5,7],
                        index=['f','g'], name='c')
        >>> ms = MulSeries([1,8,9],index=index,name=name)
        >>> ms.reset_index()
        (3, 2)    g                7
                  f                5
                    primary_index  c
        --------  --------------------
           x  y     primary_index  c
        0  1  2   0             a  1
        1  3  6   1             b  8
        2  5  6   2             b  9

        Add a col_fill:

        >>> ss_fill = pd.Series([8,9],index=['g','f'],name='primary_index'))
        >>> ms.reset_index(col_fill=ss_fill)
        (3, 2)    g             8  7
                  f             9  5
                    primary_index  c
        --------  ---------------------
           x  y     primary_index  c
        0  1  2   0             a  1
        1  3  6   1             b  8
        2  5  6   2             b  9
        '''
        if columns is None:
            if self.mindex.index.name is None:
                indexName = cmm.get_index_name('index',
                                               self.mindex.columns)
            else:
                indexName = self.mindex.index.name
            mselect = pd.DataFrame(self.mindex.index,
                                   index=self.mindex.index,
                                   columns=[indexName])
        else:
            mselect = self.mindex[columns]
            if isinstance(mselect,pd.Series):
                mselect = pd.DataFrame(mselect)
        if inplace:
            if columns is not None:
                self.mindex.drop(columns,axis=1,inplace=True)
            if not drop:
                raise TypeError('Cannot reset_index inplace on a MulSeries to create a MulDataFrame')
            if columns is None:
                self.mindex.index = range(self.shape[0])
        else:
            if columns is not None:
                mkeep = self.mindex.drop(columns,axis=1)
            else:
                mkeep = self.mindex.copy()
                mkeep.index = range(self.shape[0])
            if not drop:
                mcols = self.__fill_cols(col_fill,mselect.columns,
                                         False)
                df = pd.concat([mselect,self.ds],axis=1)
                # print(df,mcols)
                return md.MulDataFrame(df.values,index=mkeep,columns=mcols,
                              index_copy=False,columns_copy=False)
            else:
                self2 = self.copy()
                self2.index = mkeep
                return self2


    

    def call(self,func,*args,**kwargs):
        '''
        Apply a function to the values series and returns the result as a scalar or a MulSeries with the index dataframe properly arranged.

        Parameters:
        -------------
        func : function or str
            A function applied to the values series of the MulSeries. The function should return a scalar, or a pandas Series. 
            
            - If a Series is returned, it must have the same index as the primary index (order can be different if there are no duplicate values in the primary index).
            - If ``func`` is a string, it must be a valid method name of ``pandas.Series``. The method should saftisfy the same requirement as above.
        \*args : positional arguments to the function
            The MulSeries is the 1st positional argument to the function. \*args are from the 2nd positional argument onwards.
        \*\*kwargs : keyword arguments to the function
            keyword arguments to the function.

        Returns
        -----------
        scalar or MulSeries
            If the return value is a MulSeries, it should have the same index dataframe as the caller.


        Examples
        ----------
        >>> import pandas as pd
        >>> import muldataframe as md
        >>> import numpy as np
        >>> index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
        >>> name = pd.Series([5,7],
                        index=['f','g'],
                        name='c')
        >>> ms = MulSeries([1,8,9],index=index,name=name)
        >>> ms.call(np.power,2)
        (3,)      g  7
                  f  5
                     c
        -------  ------
           x  y      c
        a  1  2  a   1
        b  3  6  b  64
        b  5  6  b  81
        '''
        # if 'unsafe' in kwargs and kwargs['unsafe']:
        # consider to change to self.ss to improve safety?
        args = list(args)
        if len(args) > 0 and (isinstance(args[0],MulSeries) or \
            isinstance(args[0],md.MulDataFrame)):
            args[0] = args[0].ds
        if len(args) > 0 and hasattr(md,'__pandas_priority__') \
            and args[0].__pandas_priority__ > self.__pandas_priority__:
            return NotImplemented
        
        self.__ss._update_super_index()
        if isinstance(func,str):
            if cmm.is_pandas_method(self,func):
                res = getattr(self.__ss,func)(*args,**kwargs)
            else:
                raise ValueError(f'If func is a string, it must be a valid method name of pandas.Series')
        else:
            res = func(self.__ss,*args,**kwargs)
        errMsg = f'Currently, {self.__class__} only supports operators or functions that return a scalar value or a pandas series with the same primary index (order can be different if there are no duplicate values) in its .call() method.'
        # else:
        #     res = func(self.ss,*args,**kwargs)
        # print(res)
        if isinstance(res,pd.DataFrame):
            # print('dataframe',res)
            # have to raise error here because pandas does not support muldataframe in its __radd__ like functions.
            raise NotImplementedError(errMsg)
            # return NotImplemented
        elif isinstance(res,pd.Series):
            if res.shape[0] == self.shape[0]:
                if not res.index.equals(self.index.index):
                    try:
                        # res2 = self.loc[res.index]
                        new_idx = cmm.align_index_in_call(res.index,self,
                                                          'index')
                        if new_idx.shape[0] == self.shape[0]:
                            return MulSeries(res.values,
                                             index=new_idx,
                                             name=self.name,
                                             index_copy=False)
                        # if res2.shape[0] == self.shape[0]:
                        #     return res2
                        else:
                            raise NotImplementedError(errMsg)
                    except:
                        raise NotImplementedError(errMsg)
                else:
                    if res.name == self.name.name:
                        name = self.name.copy()
                    else:
                        name = pd.Series([],name=res.name)
                    return MulSeries(res,
                                    index=self.index,
                                    name=name,
                                    index_init='override')
            else:
                # print('shape',res.shape,self.shape)
                raise NotImplementedError(errMsg)
        else:
            return res

    def groupby(self,by=None,agg_mode:cmm.IndexAgg='same_only',keep_primary=False):
        '''
        Group MulSeries by its index dataframe using a mapper or the index dataframe's columns.

        The function uses the `DataFrame.groupby <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.groupby.html#pandas.DataFrame.groupby>`_ method of the index dataframe to create groups under the hood. The values of the MulSeries are grouped accordingly. It returns a :doc:`MulGroupBy <../groupby/indices>` object that contains information about the groups.

        Parameters
        ------------
        by : None, mapping, function, label, pd.Grouper or list of such
            Please refers to `DataFrame.groupby <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.groupby.html#pandas.DataFrame.groupby>`_ for detailed information on this argument. The difference to the :code:`by` argument in DataFrame.groupby is that if it is None, uses the primary index to group the MulSeries.
        
        agg_mode : 'same_only', 'list','tuple', default to 'same_only'
            Determine how to aggregate column values in the index dataframe that are not the same in each group when calling numpy functions on or using the :doc:`call <../groupby/indices>` method of the MulGroupBy object.

             - ``'same_only'``: only keep columns that have the same values within each group. 
             - ``'list'``: put columns that do not have the same values within a group into a list. 
             - ``'tuple'``: similar to 'list', but put them into a tuple.

        keep_primary : bool, default False
            Whether to keep primary index in the index dataframe in each group. If ``True``, the primary index will be reset as a column and kept in the grouped dataframes.  If the name of the primary index or columns is ``None``, ``"primary_index"`` will be used as its name.

        Returns
        -----------
        MulGroupBy
            A :doc:`MulGroupBy <../groupby/indices>` object that contains information about the groups.
                

        Examples
        ------------
        >>> import muldataframe as md
        >>> import pandas as pd
        >>> index = pd.DataFrame([['a','b','c'],
                                  ['g','b','f'],
                                  ['b','g','h']],
                        columns=['x','y','z'])
        >>> name = pd.Series(['a','b'],index=['e','f'],name='cc')
        >>> ms = MulSeries([1,2,3],index=index,name=name)
        >>> for key, group in ms.groupby('y'):
        ...     print(key,'\\n',group)
        ...     break
        b
        (2,)        f  b
                    e  a
                       cc
        ----------  ------
           x  y  z     cc
        0  a  b  c  0   1
        1  g  b  f  1   2
        >>> ms.groupby('y').sum()
        (2,)             f   b
                         e   a
                            cc
        ---------------  ------
        Empty DataFrame     cc
        Columns: []      y
        Index: [b, g]    b   3
                         g   3
        >>> ms.groupby('y',agg_mode='list').sum()
        (2,)               f   b
                           e   a
                              cc
        -----------------  ------
                x       z     cc
        y                  y
        b  [a, g]  [c, f]  b   3
        g       b       h  g   3
        '''
        return cmm.groupby(self,'index',by=by,
                           keep_primary=keep_primary,agg_mode=agg_mode)
    
    
    def query(self,index=None, **kwargs):
        '''
        Query the index dataframe of a MulSeries and return the the query result as a MulSeries.

        The function uses the ``pandas.DataFrame.query`` method under the hood.

        Parameters
        -----------
        index : None or str
            The query string to evaluate for the index dataframe. Check `DataFrame.query <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.query.html>`_ for detailed specification of this argument.
        kwargs : any
            The same ``kwargs`` passed to `DataFrame.query <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.query.html>`_ including the ``inplace=False`` argument.

        Returns
        ----------
        MulSeries or None
            None if ``inplace=True.``


        Examples
        ---------
        >>> index = pd.DataFrame([['a','b','c'],
                                  ['g','b','f'],
                                  ['b','g','h']],
                        columns=['x','y','z'])
        >>> name = pd.Series(['a','b'],index=['e','f'],name='cc')
        >>> ms = MulSeries([1,2,3],index=index,name=name)
        >>> ms.query('y == "b"')
        (2,)        f   b
                    e   a
                       cc
        ----------  ------
           x  y  z     cc
        0  a  b  c  0   1
        1  g  b  f  1   2
        '''
        if 'inplace' in kwargs:
            inplace = kwargs['inplace']
        else:
            inplace = False
        kwargs['inplace']=False

        if index is not None:
            rowIdx = cmm._query_index(self.mindex,index,**kwargs)
        else:
            rowIdx = slice(None)

        if not inplace:
            return self.iloc[rowIdx]
        else:
            self.__ss = self.__ss.iloc[rowIdx]
            self.index = self.index.iloc[rowIdx]
    

    def drop_duplicates(self,keep='first', inplace=False):
        '''
        Return MulSeries with duplicate values removed. 
        
        It is similar to `Series.drop_duplciates <https://pandas.pydata.org/docs/reference/api/pandas.Series.drop_duplicates.html>`_ except it returns a MulSeries with the index dataframe properly sliced.

        Parameters
        -----------
        keep : {'first', 'last', False}, defaul t 'first'
            Method to handle dropping duplicates:

            - 'first' : Drop duplicates except for the first occurrence.
            - 'last' : Drop duplicates except for the last occurrence.
            - ``False`` : Drop all duplicates.
        inplace : bool, default False
            If True, performs operation inplace and returns None.
        
        Returns:
        ----------
        MulSeries or None
            If inplace=True, returns None. Otherwise, returns a MulSeries. The MulSeries' index dataframe is properly sliced according to removed values.

        '''
        self.__ss._update_super_index()
        bidx = self.__ss.duplicated(keep=keep)
        bidx_keep = ~bidx
        new_ss = self.__ss.loc[bidx_keep]

        if inplace:
            # primary_index = self.index.index
            # primary_columns = self.columns.index
            self.__ss = None
            self.index = self.index.loc[bidx_keep]
            self.__ss = ValSeries(self,new_ss)
        else:
            return MulSeries(new_ss.values,
                        index=self.index.loc[bidx_keep],
                        name=self.name)
    
    def sort_values(self,*args,**kwargs):
        '''
        Sort by the values in the values Series.

        The input parameters are exactly the same as `pandas.Series.sort_values <https://pandas.pydata.org/docs/reference/api/pandas.Series.sort_values.html>`_ method except the return value is a mulseries.

        Parameters
        ------------
        Exactly the same as those in `pandas.Series.sort_values <https://pandas.pydata.org/docs/reference/api/pandas.Series.sort_values.html>`_.

        Returns
        ---------
        MulSeries
            A MulSeries object with the values series and the index dataframe sorted.

        Examples
        ---------
        >>> index = pd.DataFrame([['a','b','c'],
                                  ['g','b','f'],
                                  ['b','g','h']],
                        columns=['x','y','z'])
        >>> name = pd.Series(['a','b'],index=['e','f'],name='cc')
        >>> ms = MulSeries([1,2,3],index=index,name=name)
        >>> ms.sort_values(ascending=False)
        (3,)        f   b
                    e   a
                       cc
        ----------  ------
           x  y  z     cc
        2  b  g  h  2   3
        1  g  b  f  1   2
        0  a  b  c  0   1
        '''
        if self.pindex.is_unique:
            return self.call('sort_values',*args,**kwargs)
        else:
            pindex = self.pindex
            self.pindex = range(self.shape[0])
            res = self.call('sort_values',*args,**kwargs)
            res.pindex = pindex[res.pindex]
            self.pindex = pindex
            return res


ops = ['add','sub','mul','div','truediv','floordiv','mod','pow','eq','le','lt','gt','ge','ne']
for op in ops:
    op_attr = '__'+op+'__'
    def call_op_factory(op_attr):
        def call_op(self,other):
            if 'eq__' in op_attr:
                return self.equals(other)
            elif 'ne__' in op_attr:
                return not self.equals(other)
            else:
                return self.call(op_attr,other)
        return call_op
    setattr(MulSeries,op_attr,call_op_factory(op_attr))
    r_op_attr = '__r'+op+'__'
    setattr(MulSeries,r_op_attr,call_op_factory(r_op_attr))


# ops = []
# for op in ops:
#     op_attr = '__'+op+'__'
#     def call_op_factory(op_attr):
#         def call_op(self,other):
#             func = getattr(pd.Series,op_attr)
#             # print(op_attr,func)
#             return self.call(func,other)
#         return call_op
#     setattr(MulSeries,op_attr,call_op_factory(op_attr))


ValSeries = vfb.ValFrameBase_factory(pd.Series)
# def _update_super_index(self):
#     self.index = self.parent.mindex.index
#     self.name = self.parent.name.name
# ValSeries._update_super_index = _update_super_index




import numpy as np
def eq(list1, list2):
    if len(list1)!= len(list2):
        return False
    for i in range(len(list1)):
        if isinstance(list1[i],list) or isinstance(list1[i],np.ndarray):
            if not eq(list1[i],list2[i]):
                return False
        else:
            if list1[i]!= list2[i]: 
                return False
    return True


# from ..MulSeries import MulSeries
# from ..MulDataFrame import MulDataFrame
import pandas as pd
from .lib import eq
import pytest
import numpy as np
import muldataframe as md
MulDataFrame = md.MulDataFrame
MulSeries = md.MulSeries

#TODO: mloc set

def get_data():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    md = MulDataFrame([[1,2],[8,9],[8,10]],index=index,
                    columns=columns)
    return md,index,columns


def test_loc_item():
    mf, _, _ = get_data()
    mf2 = mf[mf['c']>1]
    assert eq(mf2.values,[[8,9],[8,10]])

    mf3 = mf.loc[mf['c']>1]
    assert mf2 == mf3

    mf2 = mf.loc[:,mf.loc['a']>1]
    assert eq(mf2.values,[[2],[9],[10]])

    mf3 = mf.loc[mf['c']>1,mf.loc['a']>1]
    assert eq(mf3.values,[[9],[10]])

    mf[mf['c']>1] = 5
    assert eq(mf.iloc[1:].values,[[5,5],[5,5]])

    
    # print('\n',mf.loc[mf['c']>1,mf.loc['a']>1])
    mf.loc[mf['c']>1,mf.loc['a']>1] = [[8],[9]]
    # assert eq(mf.values,[[1,2],[5,8],[5,9]])



def test_drop():
    ms, _, _ = get_data()
    ms2 = ms.copy()
    ms3 = ms2.drop('b')
    assert ms2 == ms
    assert ms3.shape == (1,2)
    assert ms3.index.shape == (1,2)

    ms2 = ms.drop(2,mloc='y')
    assert ms2.shape == (2,2)
    assert ms2.index.shape == (2,2)
    assert eq(ms2.pindex.values,['b','b'])
    assert ms2.pindex.name == ms.pindex.name
    assert ms2.mcols.shape == (2,2)

    ms.drop(5,mloc='f',inplace=True,axis=1)
    assert ms.shape == (3,1)
    assert ms.mcols.shape == (1,2)
    assert eq(ms.values,[[2],[9],[10]])
    mso, _, _ = get_data()
    assert ms.pcols.name == mso.pcols.name
    assert ms.mcols.shape == (1,2)


def test_loc():
    md, _, _ = get_data()
    md.loc['e'] = [5,7]
    assert eq(md.loc['e'].values,[5,7])
    assert md.shape == (4,2)
    assert eq(md.mindex.loc['e'].values,[None]*2)

    md, _, _ = get_data()
    md.loc['e'] = MulSeries([5,7],
        index=pd.DataFrame(index=['c','d']),
        name=pd.Series([8,9],index=['x','y']))
    # print('\n',md)
    assert eq(md.loc['e'].values,[5,7])
    assert md.shape == (4,2)
    assert eq(md.mindex.loc['e'].values,[8,9])

    md, _, _ = get_data()
    md.loc[:,'e'] = MulSeries([5,7,8],
        index=pd.DataFrame(index=['a','b','b']),
        name=pd.Series([8,9],index=['f','g']))
    # print('\n',md)
    assert eq(md.loc[:,'e'].values,[5,7,8])
    assert md.shape == (3,3)
    assert eq(md.mcols.loc['e'].values,[8,9])

    md, _, _ = get_data()
    md.loc[:,['c','d']] = [[1,2],[5,5],[6,7]]
    assert eq(md.values,[[1,2],[5,5],[6,7]])


def test_insert():
    mf,_,_ = get_data()
    mf2 = mf.copy()
    mf2.insert('e',[7,8,9])
    # print('\n',mf2)
    assert eq(mf2.pcolumns.tolist(),['c','d','e'])
    assert(mf2.shape == (3,3))
    assert eq(mf2.iloc[:,-1].values,[7,8,9])
    
    # assert eq(mf2.columns.loc['e'].values,[np.nan]*2)
    assert eq(mf2.columns.loc['e'].values,[None]*2)
    assert(mf2.mindex.shape == (3,2))

    mf2 = mf.insert('e',[7,8,9],inplace=False)
    assert eq(mf2.pcolumns.tolist(),['c','d','e'])
    assert (mf2.shape == (3,3))
    assert eq(mf2.iloc[:,-1].values,[7,8,9])

    mf2 = mf.insert('e',[7,8,9],name=[8,9],
                    inplace=False)
    assert eq(mf2.columns.loc['e'],[8,9])

    mf2 = mf.insert('e',
        MulSeries([7,8,9],
                  name=pd.Series([8,9],index=['f','g']),
                  index=pd.DataFrame(index=['a','b','b'])),
        inplace=False)
    assert eq(mf2.iloc[:,-1].values,[7,8,9])
    assert eq(mf2.columns.loc['e'],[8,9])
    mf3 = mf.insert('e',
        MulSeries([7,8,9],
                  name=pd.Series([8,9],index=['f','g']),
                  index=pd.DataFrame(index=['a','b','b'])),
        name=[10,12],
        inplace=False)
    assert(mf2 == mf3)

    mf2 = mf.insert('e',[7,8], axis=0,
                    inplace=False)
    assert mf2.shape == (4,2)
    assert mf2.index.shape == (4,2)
    assert eq(mf2.index.loc['e'],[None]*2)

    mf2 = mf.insert('e',[7,8], name=[9,10],
                    axis=0,
                    inplace=False)
    assert eq(mf2.index.loc['e'],[9,10])
    # print('\n',mf2)


def test_init():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    md = MulDataFrame([[1,2],[8,9],[8,10]],index=index,
                    columns=columns)
    
    index.iloc[0,0] = 7
    assert md.mindex.iloc[0,0] == 1
    md.mindex.iloc[2,0] = 8
    assert index.iloc[2,0] == 5
    columns.iloc[0,0] = 7
    assert md.mcolumns.iloc[0,0] == 5

    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    md = MulDataFrame([[1,2],[8,9],[8,10]],
                      index=index,
                    columns=columns,
                    both_copy=False)
    index.iloc[0,0] = 7
    assert md.mindex.iloc[0,0] == 7
    md.mindex.iloc[2,0] = 8
    assert index.iloc[2,0] == 8
    columns.iloc[0,0] = 7
    assert md.mcolumns.iloc[0,0] == 7

    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    md = MulDataFrame(index)
    assert md.mindex.shape == (3,0)
    assert md.mcolumns.shape == (2,0)

    # print('===================')
    md = MulDataFrame(index,index=index)
    assert md.index.index.equals(md.df.index)

    index2 = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['b','a','b'],
                     columns=['x','y'])
    with pytest.raises(IndexError):
        MulDataFrame(index,index=index2)
    md = MulDataFrame(index.values,index=index2)
    assert eq(md.mindex.index.values,['b','a','b'])
    md = MulDataFrame(index.values,index=index2,
                      index_init='override')
    assert eq(md.mindex.index.values,['b','a','b'])


    values = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['b','a','c'],
                     columns=['x','y'])
    index2 = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['b','a','b'],
                     columns=['x','y'])
    with pytest.raises(IndexError):
        MulDataFrame(values,index=index2)

    index2 = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['d','a','b'],
                     columns=['x','y'])
    index2.index.name = 'z'
    md = MulDataFrame(values,index=index2)
    # print(md.mindex.index.name)
    assert eq(md.mindex.index.values,['a','b'])
    assert md.mindex.index.name == 'z'

    md = MulDataFrame(values.transpose(),columns=index2)
    assert eq(md.mcols.index.values,['a','b'])
    assert md.mcols.index.name == 'z'

    columns2 = pd.DataFrame([[5,7],[3,6]],
                        index=['d','c'],
                        columns=['f','g'])
    md = MulDataFrame(columns,index=columns2,index_init='align')
    assert eq(md.mindex.index.values,['d','c'])


def test_get_set_attr():
    md,index,columns = get_data()
    assert eq(md.values,md.df.values)
    assert eq(md.values,md.ds.values)

    values = md.values
    dvals = md.df.values
    svals = md.ds.values
    values[0,0] = 5
    assert md.values[0,0] == 5
    assert svals[0,0] == 5
    assert dvals[0,0] == 1

    assert md.shape == (3,2)

    index2 = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['b','a','b'],
                     columns=['x','y'])
    md.index = index2
    assert md.df.index.equals(index2.index)
    md.mindex = index
    assert md.df.index.equals(index.index)

    # md.index = columns
    with pytest.raises(IndexError):
        md.index = columns


def test_get_setitem():
    md,index,columns = get_data()
    # print('\n',md)
    ss = md['c']
    assert ss.index.equals(md.mindex)
    assert isinstance(ss,MulSeries)
    assert eq(ss.values,[1,8,8])

    md.columns.index = ['c','c']
    md2 = md['c']
    assert md2 == md

    md,index,columns = get_data()
    md['c'] = [5,5,5]
    # print('================')
    assert eq(md.loc[:,'c'].values,[5,5,5])

    md,index,columns = get_data()
    md['e'] = [6,7,8]
    assert eq(md.loc[:,'e'].values,[6,7,8])
    assert md.shape == (3,3)
    assert eq(md.mcols.loc['e'].values,[None]*2)

    md,index,columns = get_data()
    md['e'] = MulSeries([6,7,8],
        index=pd.DataFrame(index=['a','b','b']),
        name=pd.Series([8,9],index=['f','g']))
    assert eq(md.loc[:,'e'].values,[6,7,8])
    assert md.shape == (3,3)
    # print('\n',md)
    assert eq(md.mcols.loc['e'].values,[8,9])



def test_transpose():
    md,index,columns = get_data()
    md2 = md.transpose()
    assert md2.index.equals(md.columns)
    assert md2.columns.equals(md.index)
    assert eq(md2.values.T,md.values)
    md2.iloc[0,0] = 5
    assert md.iloc[0,0] == 1
    md2.index.iloc[0,0] = 100
    assert md.columns.iloc[0,0] == 5

    md2 = md.copy()
    vals = md2.values
    index = md2.index
    md2.transpose(inplace=True)
    assert md2.index.equals(md.columns)
    assert md2.columns.equals(md.index)
    assert eq(md2.values.T,md.values)
    md2.iloc[0,0] = 5
    assert vals[0,0] == 5
    md2.columns.iloc[0,0] = 100
    assert index.iloc[0,0] == 100


def test_op_call():
    md,index,columns = get_data()
    assert md != md.df

    md2 = md*2
    md3 = 2*md
    assert md2 == md3
    assert eq(md2.iloc[:,0],[2,16,16])

    md2 = md.sum(axis=0)
    assert eq(md2.values,[17,21])

    md2 = md.power(2)
    assert md2.iloc[1,0] == 64

    md.iloc[1,0] = 6
    assert md.mean().mean() == 6

    md,index,columns = get_data()
    
    md2 = md + np.array([[1,1],[1,1],[1,1]])
    assert isinstance(md,MulDataFrame)
    assert md2.iloc[0,0] == 2
    assert md2.iloc[2,1] == 11

    md2 = md + pd.DataFrame([[1,1],[1,1],[1,1]],index=['a','b','b'],
                  columns=['c','d'])
    assert isinstance(md,MulDataFrame)
    assert md2.iloc[0,0] == 2
    assert md2.iloc[2,1] == 11

    md2 = md * pd.Series([2,2],index=['c','d'])
    assert md2.iloc[0,0] == 2
    assert md2.iloc[2,1] == 20

    md2 = md * MulSeries(pd.Series([2,2],index=['c','d']))
    assert md2.iloc[0,0] == 2
    assert md2.iloc[2,1] == 20

    md2 = md.mean(axis=1)
    assert eq(md2.values,[1.5,8.5,9.0])

    def func(df):
        return df.iloc[:2]
    with pytest.raises(NotImplementedError):
        md.call(func)
    
    def func(df):
        return df.iloc[:,[1,0]]
    md2 = md.call(func)
    assert eq(md2.iloc[0].values,[2,1])
    

def test_mloc_nloc():
    md,index,columns = get_data()
    md2 = md.mloc[[..., 6]]
    # print('\n',md.mloc[[None,6],[5]])
    assert eq(md2.values,[[8,9],[8,10]])
    md2 = md.mloc[[[1,3],6]]
    assert eq(md2.values,[8,9])

    # md.index.insert(2,'z',[1,2,3])
    # md.index.columns = ['x','y','y']
    # print('xxxxxxxxxxxx')
    # print('\n',md.mloc[{'y':2}])
    # print('\n',md.mloc[{'y':2}].name)
    md,index,columns = get_data()

    md2 = md.nloc[[..., 6]]
    assert eq(md2.values,[[8,9],[8,10]])
    md2 = md.nloc[{1:6}]
    assert eq(md2.values,[[8,9],[8,10]])


    md2 = md.mloc[[[1,3],[6]]]
    assert eq(md2.values,[[8,9]])

    with pytest.raises(KeyError):
        md2 = md.mloc[[3,6]]
    # assert eq(md2.values,[8,9])

    with pytest.raises(KeyError):
        md.mloc[[1,6]]

    md2 = md.mloc[{'y':6,'x':3}]
    assert eq(md2.values,[8,9])
    md2 = md.mloc[{'y':6,'x':[3]}]
    assert eq(md2.values,[[8,9]])
    with pytest.raises(KeyError):
        md.mloc[{'y':6,'x':[1,3]}]

    md2 = md.mloc[:,[...,7]]
    assert eq(md2.values,[1,8,8])
    md2 = md.mloc[[...,6],[...,7]]
    assert eq(md2.values,[8,8])

    md2 = md.nloc[{1:6},{1:7}]
    assert eq(md2.values,[8,8])

    md2 = md.mloc[[...,2],{'g':6}]
    # print('\n',md.mloc[:,{'g':6}])
    assert md2 == 2

    md.mloc[[...,2],{'g':6}] = 3
    assert md.iloc[0,1] == 3

    md.mloc[[...,6],{'f':5}] = [5,5]
    assert eq(md.iloc[1:,0].values, [5,5])

    md.mcols.iloc[:,0] = [5,5]
    md.mloc[[1],[5]] = [3,3]
    assert eq(md.iloc[0,:].values, [3,3])

def test_set_index():
    md,index,columns = get_data()
    md2 = md.set_index('c')
    # print('\n', md.set_index(mloc={'g':6}))
    assert md2.shape == (3,1)
    assert md2.mindex.shape == (3,3)
    assert eq(md2.mindex['c'], [1,8,8])

    md.set_index('c',inplace=True,drop=False)
    assert md.shape == (3,2)
    assert md.mindex.shape == (3,3)
    assert eq(md.mindex['c'], [1,8,8])

    md,index,columns = get_data()
    md.set_index('c',inplace=True,drop=True)
    assert md.shape == (3,1)
    assert md.mindex.shape == (3,3)
    assert eq(md.mindex['c'], [1,8,8])

    md,index,columns = get_data()
    md.mcols.iloc[:,0] = [5,5]
    md2 = md.set_index(mloc=[5])
    # print(md2)
    assert md2.shape == (3,0)
    assert md2.mindex.shape == (3,4)
    assert eq(md2.mindex['c'], [1,8,8])

    with pytest.raises(ValueError):
        md.set_index()


def test_reset_index():
    md,index,columns = get_data()
    md2 = md.reset_index()
    # print(md2)
    assert md2.mindex.shape == md.mindex.shape
    assert eq(md2.mindex.index.values,[0,1,2])
    assert eq(md2.mcols.index, ['primary_index','c','d'])
    assert eq(md2.mcols.iloc[0],['',''])
    assert md2.mcols.shape == (3,2)

    md.reset_index(inplace=True,col_fill=0)
    md2 = md
    assert md2.mindex.shape == md.mindex.shape
    assert eq(md2.mindex.index.values,[0,1,2])
    assert eq(md2.mcols.index, ['primary_index','c','d'])
    assert eq(md2.mcols.iloc[0],[0,0])
    assert md2.mcols.shape == (3,2)

    md,index,columns = get_data()
    md2 = md.reset_index(drop=True)
    assert md.mindex.shape == md2.mindex.shape
    assert md.shape == md2.shape
    assert eq(md2.index.index,[0,1,2])

    md2 = md.reset_index(['x','y'])
    # print(md,'\n',md2)
    assert md2.mindex.shape == (3,0)
    assert md2.mcols.shape == (4,2)
    assert md2.mindex.index.equals(md.mindex.index)

    md2 = md.set_index('c')
    ss = md.mcols.loc['c']
    md3 = md2.reset_index('c',col_fill=ss)
    assert md3 == md
    
    md2 = md.set_index(['c','d'])
    # print(md2)
    md3 = md2.reset_index(['c','d'],col_fill=md.mcols)
    assert md3 == md

    md3 = md2.reset_index(['c','d'],col_fill=md.mcols.loc[['d','c']])
    assert md3 == md

    with pytest.raises(IndexError):
        md2.reset_index(['c','d'],col_fill=md.mcols.loc[['d','c','c']])
    md.mcols.index = ['k','m']
    with pytest.raises(KeyError):
        md2.reset_index(['c','d'],col_fill=md.mcols)


def test_drop_duplicates():
    md,index,columns = get_data()
    md2 = md.drop_duplicates('c')
    assert eq(md2.values,[[1,2],[8,9]])

    md2 = md.drop_duplicates(['c','d'])
    assert md2.equals(md)

    md2 = md.drop_duplicates(mloc=[3])
    assert md2.equals(md)

    md2 = md.drop_duplicates(mloc={'g':7})
    assert eq(md2.values,[[1,2],[8,9]])
    # print('\n',md2)

    md.drop_duplicates('c',inplace=True)
    assert eq(md.values,[[1,2],[8,9]])

    md,index,columns = get_data()
    md2 = md.drop_duplicates()
    assert md2.equals(md)
    md.iloc[1,1] = 8
    md.iloc[2,1] = 8
    md2 = md.drop_duplicates()
    assert md2.shape == (2,2)
    assert eq(md2.values,[[1,2],[8,8]])
    

def test_iterrows():
    md,index,columns = get_data()
    for i, (k,row) in enumerate(md.iterrows()):
        # print('=============')
        # print(k,'\n',row)
        assert isinstance(row,MulSeries)
        if i == 0:
            assert k.name == 'a'
        else:
            assert k.name == 'b'


def test_iloc():
    md, index, columns = get_data()
    md2 = md.iloc[np.array([0])]
    assert eq(md2.values,[[1,2]])

def test_groupby():
    md, index, columns = get_data()
    # print('\n',md.groupby('y').mean(axis=0))
    # for key, group in md.groupby('y'):
    #     if key == 6:
    #         print(key,'\n',group)
        # break
    gpo =  md.groupby('y')
    for i, (k,gp) in enumerate(gpo):
        if i==0:
            assert k == 2
            assert gp.shape == (1,2)
            assert isinstance(gp,MulDataFrame)
        else:
            assert k == 6
            assert gp.shape == (2,2)
            assert gp.mcols.shape == (2,2)
    
    md2 = md.groupby('y').mean(axis=0)
    assert eq(md2.values,[[1,2],[8,9.5]])
    assert md2.mcols.shape == (2,2)
    assert md2.index.shape == (2,0)

    md2 = md.transpose()
    md3 = md2.groupby('y',axis=1).mean(axis=1)
    assert eq(md3.columns.index.values,[2,6])
    assert eq(md3.values,[[1,8],[2,9.5]])
    assert md3.index.shape == (2,2)

    md2 = md.groupby().mean(axis=0)
    assert eq(md2.values,[[1,2],[8,9.5]])
    assert md2.index.shape == (2,1)

    md2 = md.groupby(agg_mode='list').mean(axis=0)
    assert eq(md2.values,[[1,2],[8,9.5]])
    assert md2.index.shape == (2,2)
    # print('tttttt','\n',md)
    # print('\n',md2)
    assert eq(md2.index.iloc[1,0],[3,5])

    md2 = md.groupby(['x','y']).mean(axis=0)
    assert eq(md2.values,md.values)
    assert md2.index.shape == (3,2)
    assert eq(md2.index.index.values,[0,1,2])

    md2 = md.groupby(['x','y']).mean(axis=None)
    assert eq(md2.values,[1.5,8.5,9.0])
    assert md2.shape == (3,)
    assert md2.name.name == 'mean'

    # print('\n',md)
    mf2 = md.groupby('y').sum(axis=1)
    # print('\n',mf2)
    assert mf2.shape == (3,1)
    # print(mf2.columns.shape)
    assert mf2.columns.shape == (1,0)
    assert eq(mf2.columns.index.values, ['sum'])
    assert eq(mf2.values,[[3],[17],[18]])
    assert mf2.pindex.name == 'y'

    mf2 = md.groupby('f',axis=1).sum(axis=0)
    # print('\n',mf2)
    assert mf2.shape == (1,2)
    assert eq(mf2.values, [[21,17]])
    assert eq(mf2.pindex.values, ['sum'])
    





def test_query():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    md = MulDataFrame([[1,2],[8,9],[8,10]],index=index,
                    columns=columns)
    
    md2 = md.query('c == 8')
    assert md2.shape == (2,2)
    assert md2.iloc[0,0] == 8
    # print('\n',md2)

    md3 = md.query(index='y > 3')
    assert md3 == md2

    md4 = md.query(index='y > 3',
                   columns='f>3')
    assert md4.shape == (2,1)
    assert eq(md4.values,[[8],[8]])
    assert eq(md4.mindex.values,[[3,6],[5,6]])

    md5 = md.query('c==8',index='x<=3')
    assert md5.shape == (1,2)
    assert eq(md5.values,[[8,9]])

    md6 = md.query('c==8',index='x<=3',
                   columns='f>3')
    assert md6.shape == (1,1)
    assert eq(md6.values,[[8]])
    assert md6.mcols.shape == (1,2)
    # print('\n',md6)

    md7 = md.copy()
    md7.query('c==8',index='x<=3',
                   columns='f>3',
                   inplace=True)
    assert md6 == md7

    md8 = md.query(columns='g < 7')
    assert md8.shape == (3,1)




def test_melt():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    md = MulDataFrame([[1,2],[8,9],[8,10]],index=index,
                    columns=columns)
    
    mf = md.melt()
    # print(mf)
    # print(mf)
    assert mf.shape == (md.shape[0]*md.shape[1],7)
    assert eq(mf['value'].values,np.ravel(md.values))
    assert eq(mf.columns.tolist(),
              ['index','x','y','index','f','g','value'])
    
    mf2 = md.melt(prefix=True,value_name='num')
    assert mf2.shape == (md.shape[0]*md.shape[1],7)
    assert mf2.columns[-1] == 'num'
    assert mf2.columns[0] == 'x_index'
    assert eq(mf2['x_index'].values,
              ['a','a','b','b','b','b'])
    assert eq(mf2['y_index'].values,
              ['c','d','c','d','c','d'])
    
    md.mcols.columns = ['x','g']
    def prefix_func(indexType,label):
        if indexType == 'index':
            label = f'row_{label}'
        else:
            label = f'col_{label}'
        return label
    
    mf3 = md.melt(prefix=prefix_func,
                  ignore_primary_columns=True,
                  ignore_primary_index=True)
    assert mf3.shape == (md.shape[0]*md.shape[1],5)
    assert eq(mf3.columns.tolist(),
              ['row_x','y','col_x','g','value'])


def test_sort_values():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    mf = MulDataFrame([[1,9],[8,2],[7,10]],index=index,
                    columns=columns)
    
    mf2 = mf.sort_values('d')
    assert eq(mf2.pindex, ['b','a','b'])
    assert eq(mf2.values[:,1],[2,9,10])

    
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','c'],
                     columns=['x','y'])
    mf.index = index
    mf2 = mf.sort_values('c')
    assert eq(mf2.pindex, ['a','c','b'])
    assert eq(mf2.values[:,0], [1,7,8])

    mf2 = mf.sort_values('b',axis=1)
    # print('\n',mf2)
    assert eq(mf2.pcols, ['d','c'])
    assert eq(mf2.values[1], [2,8])

    mf.pcols = ['d','d']
    mf2 = mf.sort_values('b',axis=1)
    # print('\n',mf2)
    assert eq(mf2.pcols, ['d','d'])
    assert eq(mf2.values[1], [2,8])



    # print('\n',mf2)
    


from lib import *


attrs = ['parent','by',
         'index_agg','indexType','groupBy']

methods = ['__iter__','call']

fld = 'source/api/groupby'

generate_attr_files('MulGroupBy',
    attrs,{},'source/api/groupby',
    'cmm.MulGroupBy')

generate_method_files('MulGroupBy',
    methods,'source/api/groupby',
    'cmm.MulGroupBy')

data = [['Constructor',['mulgroupby']],
        ['Attributes',attrs],
        ['Methods',methods]]
desc = 'The class for the object returned by the :doc:`MulSeries.groupby <../mulseries/groupby>` or the :doc:`MulDataFrame.groupby,<../muldataframe/groupby>` method.'
generate_index('MulGroupBy',data,f'{fld}/indices.rst',desc)


# import cmm
# import MulSeries
# import MulDataFrame
# from . import MulSeries
# from . import MulDataFrame

# import muldataframe.MulSeries as MulSeries
from muldataframe.MulSeries import MulSeries
from muldataframe.MulDataFrame import MulDataFrame

from muldataframe.MulSeries import ValSeries
from muldataframe.util import *
# from muldataframe.ValFrameBase import ValFrameBaseMeta


# from ..MulDataFrame import MulDataFrame
# from ..MulSeries import MulSeries
# from ..util import pivot_table,concat
import pandas as pd
from .lib import eq
import muldataframe as md
MulDataFrame = md.MulDataFrame
MulSeries = md.MulSeries
from muldataframe.util import pivot_table,concat


def test_pivot_table():
    df = pd.DataFrame({"A": ["foo", "foo", "foo", "foo", "foo",
                         "bar", "bar", "bar", "bar"],
                   "B": ["one", "one", "one", "two", "two",
                         "one", "one", "two", "two"],
                   "C": ["small", "large", "large", "small",
                         "small", "large", "small", "small",
                         "large"],
                   "D": [1, 2, 2, 3, 3, 4, 5, 6, 7],
                   "E": [2, 4, 5, 5, 6, 6, 8, 9, 9]})
    
    table = pivot_table(df, values='D', index=['A', 'B'],
            columns=['C'], aggfunc="sum")
    # print('\n',table)
    assert table.index.equals(pd.DataFrame(
        [['bar','one'],['bar','two'],
         ['foo','one'],['foo','two']],
         columns=['A','B']
    ))
    assert table.columns.shape == (2,0)
    assert eq(table.columns.index,['large','small'])

def test_concat():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','c'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    mf = MulDataFrame([[1,2],[8,9],[8,10]],index=index, columns=columns)
    ms = MulSeries([5,6,7],index=pd.DataFrame(index=['a','b','c']),name=pd.Series([8,9],index=['g','f'],name='k'))
#     print('\n',mf)
#     print('\n',ms)
    mf2 = concat([mf,ms],axis=1)
    mf3 = concat([ms,mf],axis=1)
    assert (mf2.mindex.shape == (3,2))
    assert (mf3.mindex.shape == (3,0))
    assert (mf2.shape == (3,3))
    assert eq(mf2.columns.values, [[5,7],[3,6],[9,8]])
    assert eq(mf3.columns.values, [[8,9],[7,5],[6,3]])



from lib import *


ss_attrs_only = ['index','name',
         'values','ss','ds','shape',
         'primary_index','primary_name',
         'mindex','mname','pindex','pname']
ss_indexing = ['iloc','loc','mloc','nloc']

ss_attrs = ss_attrs_only+ss_indexing

# def get_lex(num):
#    # print(num)
#    fold = int(num/9)
#    residue = num%9
#    # print(fold,residue)
#    if fold > 9:
#       return get_lex(fold)+str(residue)
#    else:
#       return str(fold)+str(residue)

# mp={}
# fnames = ['00_mulseries.rst']
# for i, ax in enumerate(ss_attrs):
#    lex = get_lex(i+1)
#    fname = f'{lex}_{ax}.rst'
#    fnames.append(fname)
#    mp[ax] = fname.split('.')[0]

ss_dyn_attrs = {
   'values':
      '''
      The values of the values series. 
      
      It is not a copy.
      ''',
   'ss':
      '''
      A deep copy of the values series.
      ''',
   'ds':
      '''
      A partial copy of the values series. 

      It is different from the :doc:`MulSeries.ss <ss>` in that its values are not copied but refer to the values of the values series while its index and name are deep-copied from the values series. Use this attribute if you want to save some memory or use the same name to refer to the values series/dataframe of MulSeries/MulDataFrame.
      ''',
   'shape':
      '''
      Same as the shape of the values series.
      ''',
   'mindex':
      '''
      Alias for :doc:`MulSeries.index <index>`.
      ''',
   'mname':
      '''
      Alias for :doc:`MulSeries.name <name>`.
      ''',
   'primary_index':
      '''
      The primary index.
      
      Shorthand for ``MulSeries.index.index``.
      ''',
   'pindex':
      '''
      Alias for :doc:`MulSeries.primary_index <primary_index>`.
      ''',
   'primary_name':
      '''
      The primary name.
      
      Shorthand for ``MulSeries.name.name``.
      ''',
   'pname':
      '''
      Alias for :doc:`MulSeries.primary_name <primary_name>`.
      ''',
   
}

fnames = []
for ax in ss_attrs:
   fname = ax+'.rst'
   fnames.append(fname)
   # print(i,lex,fname)
   
   with open(f'source/api/mulseries/{fname}','w') as ff:
      underlines = '='*(len(ax)+15)
      ff.write(f'MulSeries.{ax}\n{underlines}\n\n')
      if ax not in ss_dyn_attrs:
         ff.write(f'.. autoattribute:: muldataframe.MulSeries.{ax}\n')
      else:
         ff.write('.. currentmodule:: muldataframe\n\n')
         ff.write(f'.. attribute:: MulSeries.{ax}\n')
         ff.write(ss_dyn_attrs[ax])


methods = ['__iter__','copy','equals','reset_index','query','sort_values','drop_duplicates','call','groupby','insert','drop']
method_fnames = []
for ax in methods:
   fname = ax+'.rst'
   method_fnames.append(fname)
   # print(i,lex,fname)
   
   with open(f'source/api/mulseries/{fname}','w') as ff:
      underlines = '='*(len(ax)+15)
      ff.write(f'MulSeries.{ax}\n{underlines}\n\n')
      if ax not in ss_dyn_attrs:
         ff.write(f'.. automethod:: muldataframe.MulSeries.{ax}\n')
      else:
         pass
         # ff.write('.. currentmodule:: muldataframe\n\n')
         # ff.write(f'.. attribute:: MulSeries.{ax}\n')
         # ff.write(ss_dyn_attrs[ax])


data = [['Constructor',['mulseries']],
        ['Attributes',ss_attrs_only],
        ['indexing',ss_indexing],
        ['Methods',methods]] 
generate_index('MulSeries',data,f'source/api/mulseries//indices.rst')
# with open('source/__template/index.rst','r') as rf:
#    wstr = rf.read()
#    wstr.replace('{title}','MulSeries')

#    lines = '\n'.join([f'   {x}' for x in ss_attrs_only])
#    wstr = wstr.replace('{attributes}',lines)

#    lines = '\n'.join([f'   {x}' for x in ss_indexing])
#    wstr = wstr.replace('{indexing}',lines)

#    mlines = '\n'.join([f'   {x}' for x in methods])
#    wstr = wstr.replace('{methods}',mlines)
#    with open(f'source/api/mulseries/indices.rst','w') as wf:
#       wf.write(wstr)

# # print(ss_dyn_attrs)



import pandas as pd
from .lib import eq
import pytest
import numpy as np
import muldataframe as md
MulDataFrame = md.MulDataFrame
MulSeries = md.MulSeries

def test_init():
    index = pd.DataFrame([['a','b'],['c','d']])
    with pytest.raises(IndexError):
        MulSeries([1,2,3],index=index)
    
    index = pd.DataFrame([['a','b'],['c','d']],
                         index=['e','f'])
    assert eq(MulSeries(pd.Series([1,2],index=['f','e']),
                        index=index).values,[2,1])
    assert eq(MulSeries({'f':1,'e':2},index=index,index_init='align').values,[2,1])
    assert eq(MulSeries({'f':1,'e':2,'g':3},index=index,index_init='align').values,[2,1])

    with pytest.raises(IndexError):
        MulSeries(pd.Series([1,2,3],index=['f','e','f']),
                        index=index)
    
    index = pd.DataFrame([['a','b'],['c','d'],
                          ['g','h']],
                         index=['e','f','f'])
    assert eq(MulSeries({'f':1,'e':2},index=index,index_init='align').values,[2,1,1])
    
    with pytest.raises(IndexError):
        MulSeries({'f':1,'e':2,'g':3},index=index)

def test_getattr():
    index = pd.DataFrame([['a','b'],['c','d']],
                         index=['e','f'])
    ms = MulSeries(pd.Series([1,2],index=['e','f']), 
                        index=index)

    ss = ms.ss
    ss.index = ['k','l']
    assert eq(ms.index.index.values,['e','f'])
    ss.iloc[0] = 5
    assert ms.iloc[0] == 1

    ss = ms.ds
    ss.index = ['k','l']
    assert eq(ms.index.index.values,['e','f'])
    ss.iloc[0] = 5
    assert ms.iloc[0] == 5

    assert ms.shape == (2,)

    assert ms.mindex.equals(ms.index)
    assert ms.midx.equals(ms.index)
    assert ms.mname.equals(ms.name)
    assert ms.mname.name == ms.pname


    index = pd.DataFrame([['a','b'],['c','d'],
                          ['g','h']],
                         index=['e','f','f'])
    ms = MulSeries(pd.Series([1,2,3],index=['e','f','f']), 
                        index=index)
    ms2 = ms['f']
    index = pd.DataFrame([['c','d'],
                          ['g','h']],
                         index=['f','f'])
    mse = ms = MulSeries(pd.Series([2,3],index=['f','f']), 
                        index=index)
    assert ms2 == mse



def test_setattr():
    index = pd.DataFrame([['a','b'],['c','d']],
                         index=['e','f'])
    ms = MulSeries(pd.Series([1,2],index=['e','f']), 
                        index=index)
    index = pd.DataFrame([['a','b'],
                          ['c','d'],
                          ['e','f']],
                          index=['g','h','i'])
    with pytest.raises(AttributeError):
        ms.ss = pd.Series([1,5])
    with pytest.raises(IndexError):
        ms.index = index
    

def test_ops_get_set_items():
    index = pd.DataFrame([['a','b'],['c','d']],
                         index=['e','f'])
    ms = MulSeries(pd.Series([1,3],index=['e','f']), 
                        index=index)
    ms2 = MulSeries(pd.Series([1,3],index=['e','f']), 
                        index=index)
    assert ms.equals(ms2)
    assert ms == ms2
    assert ms != pd.Series([1,3],index=['e','f'])

    # print('\n',ms)
    ms3 = ms.copy()
    ms3['g'] = 9
    assert ms3.shape == (3,)
    assert ms3['g'] == 9
    assert eq(ms3.index.loc['g'].values, [None]*2)


    def foo(ss):
        return ss.iloc[::-1]
    msf = ms.call(foo)
    assert eq(msf.values,[3,1])

    def foo2(ss):
        return ss.iloc[[0,1,1]]
    with pytest.raises(NotImplementedError):
        msf = ms.call(foo2)

    def foo3(ss):
        return pd.Series([1,3],index=['a','b'])
    with pytest.raises(NotImplementedError):
        msf = ms.call(foo3)

    ms2['e'] = 5
    assert ms != ms2
    assert ms2.ss.iloc[0] == 5
    assert ms2['e'] == 5

    ms3 = ms+ms2
    assert eq(ms3.values,[6,6])

    mse = MulSeries(pd.Series([1,3],index=['e','f']), 
                        index=index)
    ms2e = MulSeries(pd.Series([5,3],index=['e','f']), 
                        index=index)
    assert mse == ms
    assert ms2e == ms2

    msa = ms*2
    msb = 2*ms
    assert eq(msa.values,[2,6])
    assert msa == msb

    ms2 = MulSeries(pd.Series([1,3],index=['e','f']), 
                        index=index)
    ms2.index.index = ['e','k']

    assert not ms2.index.equals(index)
    with pytest.raises(NotImplementedError):
        ms + ms2
    
    index = pd.DataFrame([['a','b'],['c','d']],
                         index=['e','f'])
    ms = MulSeries(pd.Series([1,3],index=['e','f']), 
                        index=index,
                        index_copy=False)
    ms.index.index = ['e','k']
    assert eq(index.index.values,['e','k'])

    ms.index.index = ['e','f']
    msc = ms.call(np.power,2)
    assert eq(msc.values,[1,9])

    msc = ms.call(np.sum)
    assert msc == 4

    msc = ms.call(np.add,[5,6])
    # print(msc)
    assert eq(msc.values,[6,9])

    index = pd.DataFrame([['a','b'],['c','d'],
                          ['e','f']],
                         index=['a','b','b'])
    ms = MulSeries(pd.Series([1,3,5],
                    index=['a','b','b']), 
                        index=index,
                        name='abc')
    msp = ms+pd.Series([1,3,5],index=['a','b','b'])
    assert eq(msp.values,[2,6,10])
    assert msp.name.name is None

    with pytest.raises(NotImplementedError):
        ms+pd.Series([1,3,5],index=['b','a','b'])

    with pytest.raises(NotImplementedError):
        ms.call(foo)
    
    def to_df(ss):
        return pd.DataFrame([])
    
    with pytest.raises(NotImplementedError):
        ms.call(to_df)

    index = pd.DataFrame([['a','b'],['c','d']],
                         index=['e','f'])
    ms = MulSeries(pd.Series([1,3],index=['e','f']), 
                        index=index)
    assert ms.mean() == 2

    ms2 = ms.multiply(2)
    assert isinstance(ms2,MulSeries)
    assert eq(ms2.values,[2,6])

    ms2 = ms.log2()
    assert ms2.iloc[0] == 0


def test_loc():
    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','y'],
                        index=['k','l','m'])
    name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)
    
    ms2 = ms.copy()
    ms2['p'] = 9
    assert ms2.shape == (4,)
    assert eq(ms2.index.loc['p'].values, [None]*3)
    assert ms2['p'] == 9
    # print('\n',ms2)

    assert ms.loc['k'] == 1
    assert eq(ms.loc[['m','l']].values, [3,2])

    ms.index.index = ['a','b','b']
    assert eq(ms.loc['b'].values, [2,3])
    assert eq(ms.index.index.values,['a','b','b'])
    
    ms.loc['b'] = 5
    assert eq(ms.values, [1,5,5])

    ms.loc['b'] = [5,6]
    assert eq(ms.values, [1,5,6])

    ms.loc[['a','b']] = [1,2,3]
    assert eq(ms.values, [1,2,3])
    assert ms.values[-1] == 3



def test_mloc_nloc_get():
    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','y'])
    name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)
    # print(type(ms.mloc[[None,'b']].values))
    assert eq(ms.mloc[:].values,[1,2,3])

    assert eq(ms.mloc[[...,'b']].values,
                      [1,2])
    
    assert eq(ms.nloc[:].values,[1,2,3])

    assert eq(ms.nloc[[...,'b']].values,
                      [1,2])
    # print('\n',ms.mloc[[None,'b']])
    assert eq(ms.mloc[[..., ..., ['h','f']]].values,[3,2])

    with pytest.raises(KeyError):
        assert ms.mloc[['g', ..., ['h','f']]] == 2

    with pytest.raises(KeyError):
        ms.mloc[['a', ..., ['h','f']]]
    with pytest.raises(KeyError):
        ms.mloc[[['a','g'], ..., ['h','f']]]
    with pytest.raises(IndexError):
        ms.mloc[[['a','g'], ..., ['h','f'],'kk']]
    
    assert ms.mloc[['b']] == 3
    with pytest.raises(KeyError):
        ms.mloc[{'y':'b'}]

    assert eq(ms.mloc[{'y':['c','h']}].values,[1,3])
    assert ms.mloc[{'x':'b'}] == 3

    assert eq(ms.mloc[[['a','g'], ..., ['f','c']]].values,[2,1])
    assert eq(ms.mloc[[['g','a'], ..., ['c','f']]].values,[1,2])

    assert eq(ms.mloc[{'y':['c','h'],'x':['b','a']}].values,[3,1])
    # print('\n',ms.mloc[{'y':['c','h'],'x':['b','a']}])
    assert eq(ms.mloc[{'y':['h','c'],'x':['a','b']}].values,[1,3])

    assert eq(ms.nloc[{2:['c','h'],0:['b','a']}].values,[3,1])
    assert eq(ms.nloc[{2:['h','c'],0:['a','b']}].values,[1,3])


    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','z'])
    name = name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)
    assert eq(ms.mloc[[..., 'b']].values,[1,2])
    
    res = ms.mloc[{'y':[True, False, True]}]
    assert eq(res.values,[1,3])

def test_mloc_nloc_set():
    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','y'])
    name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)
    ms.mloc[[..., 'g']] = 5
    assert eq(ms.values,[1,2,5])

    ms2 = ms.copy()
    ms2.nloc[{1:'g'}] = 6
    assert eq(ms2.values,[1,2,6])
    ms2.nloc[:] = [1,2,3]

    ms.iloc[:] = [1,2,3]
    assert eq(ms.values,[1,2,3])
    ms.mloc[:] = [1,2,3]
    assert eq(ms.values,[1,2,3])

    ms.mloc[{'y':['c','f']}] = [5,6]
    assert eq(ms.values,[5,6,3])

    ms2.nloc[{2:['c','f']}] = [5,6]
    assert eq(ms2.values,[5,6,3])

    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','z'])
    name = name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)
    ms.mloc[{'y':'b'}] = [5,6]
    assert eq(ms.values,[5,6,3])


def get_data():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    ms = MulSeries([1,8,9],index=index,name=columns.loc['c'])
    return ms,index,ms.name

def test_loc_item():
    ms,_,_ = get_data()
    ms2 = ms.loc[ms>2]
    assert eq(ms2.values, [8,9])
    assert eq(ms2.pindex.values,['b','b'])

    ms3 = ms[ms>2]
    assert ms3 == ms2

    ms[ms>2] = 5
    assert eq(ms.values,[1,5,5])

    ms.loc[ms>2] = 5
    assert eq(ms.values,[1,5,5])


def test_insert():
    ms, _, _ = get_data()
    ms2 = ms.copy()
    ms2.insert('e',7)
    # print('\n',ms2)
    assert eq(ms2.pindex.tolist(),['a','b','b','e'])
    assert eq(ms2.values, [1,8,9,7])
    assert eq(ms2.mindex.iloc[-1].values,[None]*2)

    ms2 = ms.insert('e',7,1,[8,9],False)
    # print('\n',ms2)
    assert eq(ms2.pindex.tolist(),['a','e','b','b'])
    assert eq(ms2.values, [1,7,8,9])
    assert eq(ms2.mindex.iloc[1].values,[8,9])


def test_drop():
    ms, _, _ = get_data()
    ms2 = ms.copy()
    ms3 = ms2.drop('b')
    assert ms2 == ms
    assert ms3.shape == (1,)
    assert ms3.index.shape == (1,2)
    # print('\n',ms3)

    ms2.drop('a',inplace=True)
    assert ms2.shape == (2,)
    assert eq(ms2.values, [8,9])

    ms2 = ms.drop(6,mloc='y')
    assert ms2.shape == (1,)
    assert ms2.index.shape == (1,2)


def get_data2():
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                     index=['a','b','b'],
                     columns=['x','y'])
    columns = pd.DataFrame([[5,7],[3,6]],
                        index=['c','d'],
                        columns=['f','g'])
    md = MulDataFrame([[1,2],[8,9],[8,10]],index=index,
                    columns=columns)
    return md,index,columns

def test_reset_index():
    md,index,name = get_data()
    md2 = md.reset_index()
    # print('\n',md2)
    assert md2.mindex.shape == md.mindex.shape
    assert eq(md2.mindex.index.values,[0,1,2])
    assert eq(md2.mcols.index, ['primary_index','c'])
    assert eq(md2.mcols.iloc[0],['',''])
    assert md2.mcols.shape == (2,2)

    ss_fill = pd.Series(
        [8,9],index=['g','f'],name='primary_index')
    msx = md.reset_index(col_fill=ss_fill)
    # print('\n',msx)
    assert msx.columns.columns.equals(md.name.index)
    assert eq(msx.columns.loc['primary_index'].values,[9,8])


    with pytest.raises(TypeError):
        md.reset_index(inplace=True,col_fill=0)

    md,index,columns = get_data()
    md2 = md.reset_index(drop=True)
    assert md.mindex.shape == md2.mindex.shape
    assert md.shape == md2.shape
    assert eq(md2.index.index,[0,1,2])

    md2 = md.reset_index(['x','y'])
    # print(md,'\n',md2)
    assert md2.mindex.shape == (3,0)
    assert md2.mcols.shape == (3,2)
    assert md2.mindex.index.equals(md.mindex.index)
    
    md, index, columns = get_data2()
    md2 = md.set_index('c')
    ss = md.mcols.loc['c']
    ms = md2['d']
    md3 = ms.reset_index('c',col_fill=ss)
    assert md3 == md

    md2 = md.reset_index('x')
    md3 = md2.set_index(['x','c'])
    df = md2.mcols.loc[['x','c']]
    md4 = md3['d']
    md5 = md4.reset_index(['x','c'],col_fill=df)
    assert md5 == md2

    with pytest.raises(IndexError):
        md4.reset_index(['x','c'],col_fill=df.loc[['x','c','c']])
    df.index = ['k','m']
    with pytest.raises(KeyError):
        md2.reset_index(['c','d'],col_fill=df)

def test_drop_duplicates():
    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','z'])
    name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,2],index=index,name=name)

    ms2 = ms.drop_duplicates()
    assert ms2.shape == (2,)
    assert ms2.mindex.shape == (2,3)
    assert eq(ms2.mindex.index,[0,1])

    ms.drop_duplicates(inplace=True)
    ms2 = ms
    assert ms2.shape == (2,)
    assert ms2.mindex.shape == (2,3)
    assert eq(ms2.mindex.index,[0,1])


def test_query():
    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','z'])
    name = name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)
    ms2 = ms.query('y == "b"')
    assert ms2.shape == (2,)
    assert eq(ms2.values,[1,2])
    
    ms2 = ms.copy()
    ms2.query('y == "b"',inplace=True)
    assert ms2.shape == (2,)
    assert eq(ms2.values,[1,2])

    ms2 = ms.query()
    assert ms2 == ms


def test_groupby():
    index = pd.DataFrame([['a','b','c'],
                        [ 'g','b','f'],
                        [ 'b','g','h']],
                        columns=['x','y','z'])
    name = name = pd.Series(['a','b'],index=['e','f'],name='cc')
    ms = MulSeries([1,2,3],index=index,name=name)

    # for key, group in ms.groupby('y'):
    #     print(key,'\n',group)
    #     break

    # print('\n',ms.groupby('y').sum())
    # print('\n',ms.groupby('y',agg_mode='list').sum())


    assert eq(ms.groupby('y').sum().values,[3,3])
    ms2 = ms.copy()
    ms2.index.set_index('x',inplace=True)
    res = ms2.groupby('y').mean()
    assert eq(res.values,[1.5,3])
    assert eq(res.index.index.values,['b','g'])
    assert eq(res.index.columns.values,[])

    res = ms2.groupby('y',agg_mode='list').mean()
    assert eq(res.index.columns.values,['z'])
    assert eq(res.index['z'].values[0],['c','f'])

    # print('++++++++++++++++++')
    # print('\n',ms2)
    res = ms2.groupby('y',keep_primary=True,
                      agg_mode='list').mean()
    assert eq(res.index.index.values,['b','g'])
    assert eq(res.index.columns.values,['x','z'])
    assert eq(res.index['z'].values[0],['c','f'])
    assert eq(res.index['x'].values[0],['a','g'])


    def gpAdd(gp):
        if gp.mindex['y'].iloc[0] == 'b':
            return gp+5
        else:
            return gp
    msa = ms.groupby('y').call(gpAdd,use_mul=True)
    assert eq(msa.values,[6,7,3])


def test_sort_values():
    import muldataframe as md
    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                        index=['a','b','b'],
                        columns=['x','y'])
    name = pd.Series([5,7],
                            index=['f','g'],
                            name='c')
    ms = MulSeries([8,1,9],index=index,name=name)
    
    ms2 = ms.sort_values()
    # print('\n',ms2)
    assert eq(ms2.pindex,['b','a','b'])
    assert eq(ms2.values.tolist(),[1,8,9])

    index = pd.DataFrame([[1,2],[3,6],[5,6]],
                        index=['a','b','c'],
                        columns=['x','y'])
    ms.index = index
    ms2 = ms.sort_values()
    assert eq(ms2.pindex,['b','a','c'])
    assert eq(ms2.values.tolist(),[1,8,9])





def generate_attr_files(title, attrs:list[str],
    dyn_attrs:dict,fld,module=None):
    if module is None:
        module = title
    for ax in attrs:
        fname = ax+'.rst'
        with open(f'{fld}/{fname}','w') as ff:
            underlines = '='*(len(ax)+15)
            ff.write(f'{title}.{ax}\n{underlines}\n\n')
            if ax not in dyn_attrs:
                ff.write(f'.. autoattribute:: muldataframe.{module}.{ax}\n')
            else:
                ff.write('.. currentmodule:: muldataframe\n\n')
                ff.write(f'.. attribute:: {module}.{ax}\n')
                ff.write(dyn_attrs[ax])

def generate_method_files(title,methods:list[str],
                          fld,module=None):
    if module is None:
        module = title

    for ax in methods:
        fname = ax+'.rst'
        with open(f'{fld}/{fname}','w') as ff:
            underlines = '='*(len(ax)+15)
            ff.write(f'{title}.{ax}\n{underlines}\n\n')
            ff.write(f'.. automethod:: muldataframe.{module}.{ax}\n')

def generate_index(title,data,path,desc=None):
    with open(path,'w') as ff:
        underlines = '='*(len(title)+15)
        ff.write(f'{title}\n{underlines}\n\n')
        if desc:
            ff.write(f'{desc}\n\n')
        for item in data:
            desc = item[0]
            print(desc)
            lines = '\n'.join([f'   {x}' for x in item[1]])
            ff.write(f'.. toctree::\n   :maxdepth: 1\n   :caption: {desc}:\n\n')
            ff.write(lines+'\n\n')


import pandas as pd
import muldataframe.cmm as cmm
import warnings

def ValFrameBase_factory(baseClass:pd.DataFrame|pd.Series):
    class ValFrameBase(baseClass):
        def __init__(self,parent,df):
            super().__init__(df)
            # if baseClass == pd.DataFrame:
            #     print(df.shape,self.shape)
            # self.parent = parent
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.__parent = parent
            # if baseClass == pd.DataFrame:
            #     print(df.shape,self.shape)

        # @abstractmethod
        def _update_super_index(self):
            self.index = self.__parent.mindex.index
            if hasattr(self,'columns'):
                self.columns = self.__parent.mcolumns.index
            if hasattr(self,'name'):
                self.name = self.__parent.name.name

        def __getitem__(self,key):
            # # Without calling self._update_super_index(),
            # # the expression below uses self.parent.mindex.index
            # # to index but returns a dataframe or series with
            # # super(ValDataFrame,self).index.
            # res = super().__getitem__(key)
            self._update_super_index()
            return super().__getitem__(key)
        
        def __setitem__(self,key,val):
            self._update_super_index()
            return super().__setitem__(key,val)
            
        # def __getattr__(self,name):
        #     if name == 'parent':
        #         return self.__parent

        def __getattribute__(self, name:str):
            if name == 'index':
                return super().__getattribute__('__parent').mindex.index
            elif name == 'columns':
                return super().__getattribute__('__parent').mcolumns.index
            elif name == 'name':
                return super().__getattribute__('__parent').name.name
            elif name == 'iloc':
                # self._update_super_index()
                return super().__getattribute__('iloc')
            elif name == 'loc':
                
                # print(self,super().__getattribute__('columns'))
                super().__getattribute__('_update_super_index')()
                # self._update_super_index()
                return super().__getattribute__('loc')
            else:
                return super().__getattribute__(name)
    return ValFrameBase



import pandas as pd
from warnings import warn
from typing import Literal, TypeVar, Generic, Hashable
# import muldataframe.MulSeries as MulSeries
import muldataframe as md
# import muldataframe.util as util
import numpy as np
import inspect
import warnings

# MulSeries = MulSeriesModule.MulSeries

IndexType = Literal['index'] | Literal['columns']
MIndexType = Literal['mindex'] | Literal['mcolumns']
IndexInit = Literal['overlap'] | Literal['override'] | Literal['align']
IndexAgg = Literal['same_only'] | Literal['list'] | Literal['tuple']

OPS = ['add','sub','mul','div','truediv','floordiv','mod','pow']

# symbol_key = '__@$&idx'

def array_like(x):
    return isinstance(x,list) or isinstance(x,pd.Series) or isinstance(x,pd.Index) or isinstance(x,np.ndarray)

class Accessor:
    def __init__(self,getter,setter,max_idx=1):
        self.getter = getter
        self.setter = setter
        self.max_idx = max_idx
    
    def __getitem__(self,key):
        if isinstance(key,tuple) and len(key) > self.max_idx:
            raise IndexError(f"Too many indices. Only {self.max_idx} {'indices are' if self.max_idx > 1 else 'index is'} allowed")
        return self.getter(key)
    
    def __setitem__(self,key,value):
        if isinstance(key,tuple) and len(key) > self.max_idx:
            raise IndexError(f"Too many indices. Only {self.max_idx} {'indices are' if self.max_idx > 1 else 'index is'} allowed")
        # print(key,value,self.setter)
        self.setter(key,value)
    


# def _mloc_idx_each(nx,col_nx,idx):
#     if isinstance(col_nx,pd.Series):
#         col_nx_reverse = pd.Series(col_nx.index.values,
#                                    index=col_nx.values)
#         col_idx = col_nx_reverse.loc[idx]
#         return col_idx.values if isinstance(col_idx,pd.Series) \
#               else col_idx
#     else:
#         err_no_overlap = 'The input indices to the multi-index dataframe do not overlap.'
#         if isinstance(idx,list):
#             if col_nx in idx:
#                 return nx
#             else:
#                 raise KeyError(err_no_overlap)
#         else:
#             if col_nx == idx:
#                 return nx
#             else:
#                 raise KeyError(err_no_overlap)


# def _mloc_idx_old(key,df):
#     nx = list(range(df.shape[0]))
#     if isinstance(key,list):
#         if len(key) > df.shape[0]:
#             raise IndexError(f'Too many indices. There should be at most {df.shape[0]} indices.')
#         for i, idx in enumerate(key):
#             if idx is not ...:
#                 ss = df.iloc[:,i]
#                 ss.index = list(range(df.shape[0]))
#                 col_nx = ss.loc[nx]
#                 nx = _mloc_idx_each(nx,col_nx,idx)
#     else:
#         for k, idx in key.items():
#             colx = df.loc[:,k]
#             if isinstance(colx, pd.DataFrame):
#                 colx = colx.iloc[:,-1]
#                 msg = f'There are more multiple {k} columns in index dataframe. Use only the last {k} column.'
#                 warn(msg)
#             colx.index = list(range(df.shape[0]))
#             col_nx = colx.loc[nx]
#             nx = _mloc_idx_each(nx,col_nx,idx)
#     return nx


def _sucessive_indexing(ss,idx_arr):
    drop_level_count = 0
    is_scalar = False
    for i,idx_item in enumerate(idx_arr):
        if idx_item == slice(None):
            continue
        if is_scalar:
            #raise error is there are idx_items after scalar selection
            # After scalar selection, there should not be any other idx_items
            raise KeyError(f'There can be no multiple-index selection after scalar selection.')
            # if idx_item == slice(None):
            #     continue
            # item = ss.index.get_level_values(i-drop_level_count)[0]
            # # print(idx_item)
            # err_msg = 'The input indices to the multi-index dataframe do not overlap.'
            # if isinstance(idx_item,Hashable):
            #     if item != idx_item:
            #         raise KeyError(err_msg)
            # else:
            #     if item not in idx_item:
            #         raise KeyError(err_msg)
        else:
            idx = [slice(None)]*(i-drop_level_count)+[idx_item]
            level_count = len(ss.index.names)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                ss = ss.loc[tuple(idx)]
            if not isinstance(ss,pd.Series):
                is_scalar = True
                break
            if ss.shape[0] == 1 and isinstance(idx_item,Hashable):
                is_scalar = True
            if len(ss.index.names) < level_count:
                drop_level_count += 1
    # print('=======',ss)
    if is_scalar:
        return ss.values[0] if isinstance(ss,pd.Series) else ss
    else:
        return ss.values

# def _mloc_idx2(key,df):
#     original_cols = df.columns
#     df.columns = list(range(df.shape[1]))
#     num_col = object()
#     df[num_col] = list(range(df.shape[0]))
#     original_idx = df.index
#     first = True
#     for i, idx in enumerate(key):
#         # print(i,idx)
#         if idx is ... or idx == slice(None):
#             continue
#         if first:
#             df.set_index(i,drop=False,inplace=True)
#             df2 = df
#             # print(df)
#             # print(df2)
#             first = False
#         else:
#             df2.set_index(i,inplace=True)
#         df2 = df2.loc[idx]
#     del df[num_col]
#     df.columns = original_cols
#     df.index = original_idx
#     return df2[num_col].values

def _mloc_idx(key,df):
    nx = list(range(df.shape[0]))
    if isinstance(key,list):
        if len(key) > df.shape[0]:
            raise IndexError(f'Too many indices. There should be at most {df.shape[0]} indices.')
        idx_arr = [x if x is not ... else slice(None) for x in key]
        ss = pd.Series(nx,index=pd.MultiIndex.from_frame(df),copy=False)
        # if len(idx_arr) < df.shape[1]:
        #     df2 = df.iloc[:,:len(idx_arr)]
        # else:
        #     df2 = df.copy()
    else:
        idx_arr = key.values()
        sk = pd.Series(range(df.shape[1]),index=df.columns)
        idx_num = []
        for k in key.keys():
            nk = sk[k]
            if isinstance(nk,pd.Series):
                nk = nk.iloc[-1]
                msg = f'There are more multiple {k} columns in index dataframe. Use only the last {k} column.'
                warn(msg)
            idx_num.append(nk)
        df2 = df.iloc[:,idx_num]
        ss = pd.Series(nx,index=pd.MultiIndex.from_frame(df2),copy=False)
    return _sucessive_indexing(ss,idx_arr)
    
    # if isinstance(ss2,pd.Series) and ss2.shape[0] == 1 and not hasattr(idx_arr[-1], "__len__"):
    #     return ss2.values[0]
    # else:
    #     return ss2.values
    
    # ss2 = ss.loc[tuple(idx_arr)]
    # return ss2.values



def _nloc_idx(key:dict,df):
    nx = list(range(df.shape[0]))
    idx_arr = list(key.values())
    df2 = df.iloc[:,list(key.keys())]
    ss = pd.Series(nx,index=pd.MultiIndex.from_frame(df2),copy=False)
    return _sucessive_indexing(ss,idx_arr)
    # ss2 = ss.loc[tuple(idx_arr)]
    # print(ss2)
    # if isinstance(ss2,pd.Series) and ss2.shape[0] == 1 and not hasattr(idx_arr[-1], "__len__"):
    #     return ss2.values[0]
    # else:
    #     return ss2.values


# def _nloc_idx_old(key:dict,df):
#     nx = list(range(df.shape[0]))
#     for k, idx in key.items():
#         colx = df.iloc[:,k]
#         colx.index = list(range(df.shape[0]))
#         col_nx = colx.loc[nx]
#         nx = _mloc_idx_each(nx,col_nx,idx)
#     return nx


# def _n2m_key(df,key:dict):
#     key2 = {}
#     max_pos = 0
#     for k,v in key.items():
#         if not isinstance(k,int):
#             raise ValueError(f"The keys of a dict in nloc indexing can only be integers.{k} found.")
#         pos = df.shape[1]+k if k<0 else k
#         key2[pos] = v
#         if max_pos < pos:
#             max_pos = pos
#     if max_pos >= df.shape[1]:
#         raise ValueError("The position index must be smaller than the index dataframe's columns' length")
#     key3 = [...]*(max_pos+1)
#     for k,v in key2.items():
#         key3[k] = v
#     return key3

def _query_index(df,expr,**kwargs):
    # col = '__@$&idx'
    col = object() # unique key
    df[col] = list(range(df.shape[0]))
    df2 = df.query(expr,**kwargs)
    del df[col]
    return df2[col].tolist()

def checkAlign(label,ss_shape,index_shape):
    if ss_shape > index_shape:
        raise IndexError(f'Index Align: The {label} of the values Series/dataframe is not unique.')
    
def checkOverride(label,ss_shape,index_shape):
    if ss_shape != index_shape:
        raise IndexError(f'Index Overriding: The index length of the {label} dataframe is not the same as the {label} length of the values dataframe.')

def checkOverlap(label,valIndex,index):
    if not valIndex.is_unique or not index.is_unique:
        raise IndexError(f'Index Overlap: The {label} of the values dataframe and the index oof the {label} dataframe must be unique.')


def setMulIndex(dx:pd.Series|pd.DataFrame,indexType:IndexType,
                index:pd.DataFrame|None, index_init:IndexInit,
                index_copy:bool):
    if index is None:
        index = pd.DataFrame([],index=getattr(dx,indexType))
    else:
        if index_copy:
            index = index.copy()
        if index_init == 'override':
            checkOverride(indexType,getattr(dx,indexType).shape[0],
                          index.shape[0])
            # setattr(dx,indexType,index.index)
        elif index_init == 'overlap':
            if not index.index.equals(getattr(dx,indexType)):
                checkOverlap(indexType,getattr(dx,indexType),index.index)
                overlap = index.index.intersection(getattr(dx,indexType)).tolist()
                dx = dx.loc[overlap] if indexType == 'index' else dx.loc[:,overlap]
                index = index.loc[overlap]
        else:
            # print(index,getattr(dx,indexType))
            if not index.index.equals(getattr(dx,indexType)):
                if indexType == 'index':
                    dx = dx.loc[index.index]
                else:
                    dx = dx.loc[:,index.index]
                checkAlign(indexType,getattr(dx,indexType).shape[0],index.shape[0])
    return dx, index


def checkSetIdxValue(self,name,value):
    className = self.__class__.__name__
    shapeIdx = 0 if name == 'index' else 1
    if not self._hasVal():
            return
    if not isinstance(value,pd.DataFrame) or \
        value.shape[0] != self.shape[shapeIdx]:
        raise IndexError(f"The assigned value must be a dataframe with its index length being the same as the {className}'s {name} length.")

def test_idx_eq(mindex,idx,indexType='index',
                copy=True,
                err1=KeyError('Failed to index the dataframe "mindex" using "idx"'),
                err2=IndexError('The indexed new dataframe does not have the same shape as the original "mindex" dataframe. Possibly there are duplicate values in the index of the "mindex" dataframe.')):
    # equal index even if order is not the same.
    if getattr(mindex,indexType).equals(idx):
        return mindex.copy() if copy else mindex
    try:
        if indexType == 'index':
            new_idx = mindex.loc[idx]
            if new_idx.shape[0] == mindex.shape[0]:
                return new_idx
            else:
                raise err2
        else:
            new_idx = mindex.loc[:,idx]
            if new_idx.shape[1] == mindex.shape[1]:
                return new_idx
            else:
                raise err2
    except:
        raise err1

def align_index_in_call(idx,self,indexType:IndexType):
    return test_idx_eq(getattr(self,indexType),idx,
                       err1=NotImplementedError,
                       err2=NotImplementedError)


def get_index_name(indexType,arr):
    for i in range(1000):
        name = f'primary_{indexType}' if i==0 else f'primary_{indexType}_{i}'
        if name not in arr:
            return name

def is_pandas_method(self,name):
        if isinstance(self,md.MulSeries):
            pdClass = pd.Series
        else:
            pdClass = pd.DataFrame
        return hasattr(pdClass,name) and hasattr(getattr(pdClass,name),'__call__')
    
def is_numpy_function(name):
    return hasattr(np,name) and hasattr(getattr(np,name),'__call__')


def groupby(self,indexType:IndexType|MIndexType,by=None,
            keep_primary=False,
            agg_mode:IndexAgg='same_only',
            **kwargs):
        index_agg = agg_mode
        if isinstance(self,md.MulSeries):
            M = md.MulSeries
        else:
            M = md.MulDataFrame
        if by is None or (isinstance(by,list) and None in by) or keep_primary:
            # # .loc[:] does not return a copy!!!
            # ms = self.loc[:]
            ms = self.copy(data_copy=False)
            if getattr(self,indexType).index.name is None:
                getattr(ms,indexType).index.name = \
                    get_index_name('index',getattr(self,indexType).columns)
            primary_name = getattr(ms,indexType).index.name
            setattr(ms,indexType,getattr(ms,indexType).reset_index())
            # ms.index = self.index.reset_index()
            if by is None:
                by = primary_name
            elif isinstance(by,list) and None in by:
                by = [primary_name if b is None else b for b in by]
            groupBy = getattr(ms,indexType).groupby(by,**kwargs)
            return MulGroupBy[M](ms,indexType,by,groupBy,index_agg)
        else:
            groupBy = getattr(self,indexType).groupby(by,**kwargs)
            return MulGroupBy[M](self,indexType,by,groupBy,index_agg)

# G = TypeVar('G')
M = TypeVar('M')

class MulGroupBy(Generic[M]):
    def __init__(self,parent:M,indexType:IndexType|MIndexType,
                 by, groupBy:pd.core.groupby.DataFrameGroupBy,
                 index_agg:IndexAgg):
        self.groupBy = groupBy
        '''
        A `pandas.api.typing.DataFrameGroupBy <https://pandas.pydata.org/docs/reference/groupby.html>`_ object.
        '''
        self.parent = parent
        '''
        The parent MulSeries or MulDataFrame that calls the grouby method.
        '''
        self.by = by
        '''
        Same as the ``by`` argument in the parent's groupby method
        '''
        self.index_agg = index_agg
        '''
        Same as the by ``agg_mode`` argument in the parent's groupby method
        '''
        self.indexType = indexType
        '''
        The index dataframe used to group by the parent.

        It must be ``'index'`` if parent is a MulSeries. It can be ``'index'`` or ``'columns'`` if parent is a MulDataFrame.
        '''
    
    def __iter__(self):
        '''
        Make the MulGroupBy object iterable.
        '''
        if self.indexType in ['index','mindex','midx']:
            for k,v in self.groupBy.indices.items():
                yield k, self.parent.iloc[v]
        else:
            for k,v in self.groupBy.indices.items():
                yield k, self.parent.iloc[:,v]
    

    def __getattr__(self,name):
        if hasattr(np,name) and hasattr(getattr(np,name),'__call__'):
            np_func = getattr(np,name)
            params = inspect.signature(np_func).parameters
            def func(*args,**kwargs):
                if 'axis' in params and 'axis' not in kwargs:
                    kwargs['axis'] = 0 if self.indexType == 'index' else 1
                return self.call(np_func,*args,**kwargs)
            return func
        
    def call(self,func,*args,
             use_mul=False,
             set_primary=True,
             **kwargs):
        '''
        Call a function on the MulGroupBy object.

        Similar to `MulSeries.call <../mulseries/call>`_ and `MulDataFrame.call <../muldataframe/call>`_, it applies a function to the MulSeries or MulDataFrame in each group and concateate the results into a final MulSeries or MulDataFrame.

        Parameters:
        ------------
        func: function
            A function applied to the MulSeries or MulDataFrame in each group.
        use_mul: bool, default False
            An optional argument to determine how :code:`func` is applied. If False, :code:`MulSeries.call(func)` or :code:`MulDataFrame.call(func)` is used to compute the results for each group. The object passed to :code:`func` will be the MulSeries or the MulDataframe's values Series or DataFrame in each group. If True, :code:`func(MulSeries)` or :code:`func(MulDataFrame)` are used to compute the results.
        set_primary : bool, default True
            This parameter will only work if ``self.by`` is ``None`` or is a column label that selects a column in the parent's index or columns dataframe. It sets the values in the ``by`` column as the primary index or columns of the return value.
        
        Returns
        --------
        MulSeries or MulDataFrame
            The return value is a MulSeries if the MulGroupBy object's parent is a MulSeries. Otherwise, it is a MulDataFrame
        '''
        # print('******',G,M,type(G),)
        M_class = self.__orig_class__.__args__[0]
        res = None
        arr = []
        for i,(key,gp) in enumerate(self):
            # print(key,gp)
            if use_mul:
                val = func(gp,*args,**kwargs)
            else:
                val = gp.call(func,*args,**kwargs)
            # print('\n',gp,'\n',val)
            if isinstance(val,M_class):
               arr.append(val)
                # return NotImplemented
            else:
                index = gp.index if self.indexType == 'index' else gp.columns
                # print(index)
                index = md.aggregate_index(i,index,self.index_agg)
                # print('\n',index)
                if not isinstance(val,md.MulDataFrame) and \
                    not isinstance(val,md.MulSeries):
                    if isinstance(self.parent,md.MulSeries):
                        name = self.parent.name.copy()
                    else:
                        name = func.__name__
                    ms = md.MulSeries([val],index=index,
                                name=name)
                elif isinstance(self.parent,md.MulDataFrame) and \
                    isinstance(val,md.MulSeries):
                    if self.indexType == 'index':
                        if kwargs['axis'] == 0:
                            ms = md.MulDataFrame([val.values],
                                index=index,
                                columns=val.index,both_copy=False)
                        else:
                            ms = md.MulDataFrame([[x] for x in val.values],
                                index=val.index,
                                columns=pd.DataFrame(val.name).transpose(),both_copy=False)
                        #  print(val,ms)
                    else:
                        if kwargs['axis'] == 1:
                            ms = md.MulDataFrame([[x] for x in val.values],
                                index=val.index,
                                columns=index,both_copy=False)
                        else:
                            ms = md.MulDataFrame([val.values],
                                index=pd.DataFrame(val.name).transpose(),
                                columns=val.index,both_copy=False)
                        # print(val,'kkkkk\n',ms)
                else:
                    raise NotImplementedError('The function applied to a group MulDataFrame can only produce a A MulDataFrame, a MulSeries or a scalar. The function applied to a group MulSeries can only produce a MulSeries or a scalar.')
                arr.append(ms)

        axis = 0 if self.indexType == 'index' or \
             isinstance(arr[0],md.MulSeries) else 1
        # print('res','\n',arr[0],'\n',arr[1])
        res = md.concat(arr,axis=axis)
        if isinstance(self.by,str) and set_primary:
            if self.indexType == 'index':
                res.index.set_index(self.by,inplace=True)
            else:
                res.columns.set_index(self.by,inplace=True)
        return res

def fmtSeries(ss:pd.Series):
    if ss.shape[0] == 0:
        return ss
    else:
        df = pd.DataFrame(ss)
        if ss.name is None:
            df.columns = [None]
        return df


def fmtColStr(df:pd.DataFrame,transpose=True):
    if df.shape[0] == 0 or df.shape[1] == 0:
        return str(df)
    else:
        if transpose:
            df = df.transpose()
        xs = str(df)
        lines = xs.split('\n')
        lines = lines[1:][::-1]+[lines[0]]
        return '\n'.join(lines)


# funcs = ['mean','median','std','var','sum','prod','count','first','last','mad']
# funcs = ['mean','median','std','var','sum','prod']
# for func_name in funcs:
#     def call_func_factory(func_name):
#         def call_func(self,*args,**kwargs):
#             func = getattr(np,func_name)
#             # print(op_attr,func)
#             return self.call(func,*args,**kwargs)
#         return call_func
#     setattr(MulGroupBy,func_name,call_func_factory(func_name))


# def concat(ss1:MulSeries,ss2:MulSeries):
#     ss_new = ss1.ss.concat(ss2.ss)
#     index_new = pd.concat([ss1.index,ss2.index],join='inner')
#     return MulSeries(ss_new,index=index_new,
#                      name=ss1.name.copy)



from lib import *

fld = 'source/api/muldataframe'
name = 'MulDataFrame'


attrs_only = ['index','columns',
         'values','df','ds','shape',
         'primary_index', 'primary_columns',
         'mindex','mcolumns','mcols','pindex','pcolumns','pcols']
attrs_indexing = ['iloc','loc','mloc','nloc']

dyn_attrs = {
   'values':
      '''
      The values of the values dataframe.
    
      It is not a copy.
      ''',
   'df':
      '''
      A deep copy of the values dataframe.
      ''',
   'ds':
      '''
      A partial copy of the values dataframe. 

      It is different from the :doc:`MulDataFrame.df <df>` in that its values are not copied but refer to the values of the values dataframe while its index and columns are deep-copied from the values dataframe. Use this attribute if you want to save some memory or use the same name to refer to the values series/dataframe of MulSeries/MulDataFrame.
      ''',
   'shape':
      '''
      Same as the shape of the values dataframe.
      ''',
   'mindex':
      '''
      Alias for :doc:`MulDataFrame.index <index>`.
      ''',
   'mcolumns':
      '''
      Alias for :doc:`MulDataFrame.columns <columns>`.
      ''',
   'mcols':
      '''
      Alias for :doc:`MulDataFrame.columns <columns>`.
      ''',
   'primary_index':
      '''
      The primary index.
      
      Shorthand for ``MulDataFrame.index.index``.
      ''',
   'pindex':
      '''
      Alias for :doc:`MulDataFrame.primary_index <primary_index>`.
      ''',
   'primary_columns':
      '''
      The primary columns.
      
      Shorthand for ``MulDataFrame.columns.index``.
      ''',
   'pcolumns':
      '''
      Alias for :doc:`MulDataFrame.primary_columns <primary_columns>`.
      ''',
   'pcols':
      '''
      Alias for :doc:`MulDataFrame.primary_columns <primary_columns>`.
      ''',
}


generate_attr_files(name,
    attrs_only+attrs_indexing,dyn_attrs,fld)


methods = ['__iter__','copy','equals','transpose',
'set_index','reset_index','query','iterrows',
'sort_values','drop_duplicates','call',
'groupby','melt',
'insert','drop']

generate_method_files(name,
    methods,fld)

data = [['Constructor',['muldataframe']],
        ['Attributes',attrs_only],
        ['Indexing',attrs_indexing],
        ['Methods',methods]]
generate_index(name,data,f'{fld}/indices.rst')


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import pathlib
import sys
# print(pathlib.Path(__file__).parents[2])
fld = pathlib.Path(__file__).parents[2].resolve().as_posix()
sys.path.insert(0,fld)

project = 'MulDataFrame'
copyright = '2024, Qiaonan Duan and Fei Wang'
author = 'Qiaonan Duan, Fei Wang'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'myst_parser',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx_favicon',
    # 'sphinx_sitemap'
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_title = '<img style="width:50px; margin-right:0.5em" src="https://frlender.github.io/muldataframe-doc/_static/android-chrome-192x192.png">MulDataFrame'
html_theme = 'pydata_sphinx_theme'
# html_title_format = '{title}'
html_static_path = ['_static']

# favicons = [
#     "favicon-16x16.png",
#     "favicon-32x32.png",
#     "favicon.ico",
# ]

favicons = [
    # "https://frlender.github.io/muldataframe-doc/_static/favicon-16x16.png",
    # "https://frlender.github.io/muldataframe-doc/_static/favicon-32x32.png",
    "https://frlender.github.io/muldataframe-doc/_static/favicon.ico?",
]